import logging
import random

import patronus
from patronus import evaluator, Client, task, retry

cli = Client()

formatter = logging.Formatter("[%(levelname)-5s] [%(name)-10s] %(message)s")
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.DEBUG)
console_handler.setFormatter(formatter)

plog = logging.getLogger("patronus")
plog.setLevel(logging.DEBUG)
plog.propagate = False
plog.addHandler(console_handler)


@task
@retry(max_attempts=3)
async def say_hi(evaluated_model_input: str) -> str:
    r = random.random()
    # if r < 0.5:
    #     raise Exception(f"Task random exception; r={r}")
    return f"Hi {evaluated_model_input}"


@evaluator(criteria="123")
# @retry(max_attempts=3)
def iexact_match(evaluated_model_output: str, evaluated_model_gold_answer: str) -> bool:
    # async def iexact_match(evaluated_model_output: str, evaluated_model_gold_answer: str) -> bool:
    r = random.random()
    # if r < 0.5:
    #     raise Exception(f"Evaluation random exception; r={r}")
    # await asyncio.sleep(random.random() * 3)
    return (
        evaluated_model_output.lower().strip()
        == evaluated_model_gold_answer.lower().strip()
    )


@evaluator
def rnd(row) -> patronus.EvaluationResult:
    r = random.random()
    pass_ = r > 0.5
    return patronus.EvaluationResult(pass_=pass_, score_raw=r)


entry = {
    "evaluated_model_input": "Bar",
    "evaluated_model_gold_answer": "eloh Bar!",
}

ex = cli.experiment(
    "local-imatch-evals",
    data=[entry] * 10,
    task=say_hi,
    evaluators=[iexact_match, rnd],
    max_concurrency=5,
)



"""
Context management for Patronus SDK.

This module provides classes and utility functions for managing the global Patronus
context and accessing different components of the SDK like logging, tracing, and API clients.
"""

import dataclasses
import logging
from opentelemetry import trace
from typing import Optional
from typing import TYPE_CHECKING

from patronus.context.context_utils import ContextObject
from patronus.exceptions import UninitializedError

if TYPE_CHECKING:
    from patronus.evals.exporter import BatchEvaluationExporter
    from patronus.tracing.logger import Logger as PatLogger
    from patronus.api import PatronusAPIClient
    from opentelemetry.sdk.trace import TracerProvider
    from opentelemetry._logs import LoggerProvider


@dataclasses.dataclass(frozen=True)
class PatronusScope:
    """
    Scope information for Patronus context.

    Defines the scope of the current Patronus application or experiment.

    Attributes:
        service: The service name as defined in OTeL.
        project_name: The project name.
        app: The application name.
        experiment_id: The unique identifier for the experiment.
        experiment_name: The name of the experiment.
    """

    service: Optional[str]
    project_name: Optional[str]
    app: Optional[str]
    experiment_id: Optional[str]
    experiment_name: Optional[str]


@dataclasses.dataclass(frozen=True)
class PatronusContext:
    """
    Context object for Patronus SDK.

    Contains all the necessary components for the SDK to function properly.

    Attributes:
        scope: Scope information for this context.
        tracer_provider: The OpenTelemetry tracer provider.
        logger_provider: The OpenTelemetry logger provider.
        api_client: Client for Patronus API communication.
        exporter: Exporter for batch evaluation results.
    """

    scope: PatronusScope
    tracer_provider: "TracerProvider"
    logger_provider: "LoggerProvider"
    api_client: "PatronusAPIClient"
    exporter: "BatchEvaluationExporter"


_CTX_PAT = ContextObject[PatronusContext]("ctx.pat")


def set_global_patronus_context(ctx: PatronusContext):
    """
    Set the global Patronus context.

    Args:
        ctx: The Patronus context to set globally.
    """
    _CTX_PAT.set_global(ctx)


def get_current_context_or_none() -> Optional[PatronusContext]:
    """
    Get the current Patronus context or None if not initialized.

    Returns:
        The current PatronusContext if set, otherwise None.
    """
    return _CTX_PAT.get()


def get_current_context() -> PatronusContext:
    """
    Get the current Patronus context.

    Returns:
        The current PatronusContext.

    Raises:
        UninitializedError: If no active Patronus context is found.
    """
    ctx = get_current_context_or_none()
    if ctx is None:
        raise UninitializedError(
            "No active Patronus context found. Please initialize the library by calling patronus.init()."
        )
    return ctx


def get_logger(ctx: Optional[PatronusContext] = None, level: int = logging.INFO) -> logging.Logger:
    """
    Get a standard Python logger configured with the Patronus context.

    Args:
        ctx: The Patronus context to use. If None, uses the current context.
        level: The logging level to set. Defaults to INFO.

    Returns:
        A configured Python logger.
    """
    from patronus.tracing.logger import set_logger_handler

    ctx = ctx or get_current_context()

    logger = logging.getLogger("patronus.sdk")
    set_logger_handler(logger, ctx.scope, ctx.logger_provider)
    logger.setLevel(level)
    return logger


def get_logger_or_none(level: int = logging.INFO) -> Optional[logging.Logger]:
    """
    Get a standard Python logger or None if context is not initialized.

    Args:
        level: The logging level to set. Defaults to INFO.

    Returns:
        A configured Python logger if context is available, otherwise None.
    """
    ctx = get_current_context()
    if ctx is None:
        return None
    return get_logger(ctx, level=level)


def get_pat_logger(ctx: Optional[PatronusContext] = None) -> "PatLogger":
    """
    Get a Patronus logger.

    Args:
        ctx: The Patronus context to use. If None, uses the current context.

    Returns:
        A Patronus logger.
    """
    ctx = ctx or get_current_context()
    return ctx.logger_provider.get_logger("patronus.sdk")


def get_pat_logger_or_none() -> Optional["PatLogger"]:
    """
    Get a Patronus logger or None if context is not initialized.

    Returns:
        A Patronus logger if context is available, otherwise None.
    """
    ctx = get_current_context_or_none()
    if ctx is None:
        return None

    return ctx.logger_provider.get_logger("patronus.sdk")


def get_tracer(ctx: Optional[PatronusContext] = None) -> trace.Tracer:
    """
    Get an OpenTelemetry tracer.

    Args:
        ctx: The Patronus context to use. If None, uses the current context.

    Returns:
        An OpenTelemetry tracer.
    """
    ctx = ctx or get_current_context()
    return ctx.tracer_provider.get_tracer("patronus.sdk")


def get_tracer_or_none() -> Optional[trace.Tracer]:
    """
    Get an OpenTelemetry tracer or None if context is not initialized.

    Returns:
        An OpenTelemetry tracer if context is available, otherwise None.
    """
    ctx = get_current_context_or_none()
    if ctx is None:
        return None
    return ctx.tracer_provider.get_tracer("patronus.sdk")


def get_api_client(ctx: Optional[PatronusContext] = None) -> "PatronusAPIClient":
    """
    Get the Patronus API client.

    Args:
        ctx: The Patronus context to use. If None, uses the current context.

    Returns:
        The Patronus API client.
    """
    ctx = ctx or get_current_context()
    return ctx.api_client


def get_api_client_or_none() -> Optional["PatronusAPIClient"]:
    """
    Get the Patronus API client or None if context is not initialized.

    Returns:
        The Patronus API client if context is available, otherwise None.
    """
    return (ctx := get_current_context_or_none()) and ctx.api_client


def get_exporter(ctx: Optional[PatronusContext] = None) -> "BatchEvaluationExporter":
    """
    Get the batch evaluation exporter.

    Args:
        ctx: The Patronus context to use. If None, uses the current context.

    Returns:
        The batch evaluation exporter.
    """
    ctx = ctx or get_current_context()
    return ctx.exporter


def get_exporter_or_none() -> Optional["BatchEvaluationExporter"]:
    """
    Get the batch evaluation exporter or None if context is not initialized.

    Returns:
        The batch evaluation exporter if context is available, otherwise None.
    """
    return (ctx := get_current_context_or_none()) and ctx.exporter


def get_scope(ctx: Optional[PatronusContext] = None) -> PatronusScope:
    """
    Get the Patronus scope.

    Args:
        ctx: The Patronus context to use. If None, uses the current context.

    Returns:
        The Patronus scope.
    """
    ctx = ctx or get_current_context()
    return ctx.scope


def get_scope_or_none() -> Optional[PatronusScope]:
    """
    Get the Patronus scope or None if context is not initialized.

    Returns:
        The Patronus scope if context is available, otherwise None.
    """
    return (ctx := get_current_context_or_none()) and ctx.scope



import logging
import typing

from openai import OpenAI

import patronus
from patronus.evaluators_remote import EvaluateCall

cli = patronus.Client()
oai = OpenAI()

dataset = [
    {
        "model": "gpt-4o-mini",
        "input": (
            "Write a hello world in Python language. "
            "Make sure that the output is a valid python code. "
            "The python cannot be a pythonic, well written code. "
            "It should be done in very unpythonic way."
        ),
    },
    {
        "model": "gpt-4o",
        "input": "Write hello world in Java",
    },
]


@patronus.task
def call_llm(row: patronus.Row) -> patronus.TaskResult:
    model_output = (
        oai.chat.completions.create(
            model=row.model,
            messages=[
                {"role": "user", "content": row.input},
            ],
            temperature=0,
        )
        .choices[0]
        .message.content
    )
    return patronus.TaskResult(
        evaluated_model_output=model_output,
        metadata={"model": row.model},
        tags={"model": row.model},
    )


base_eval_python_code = cli.remote_evaluator(
    "custom-small",
    criteria="is-python-code",
    criteria_config={
        "pass_criteria": """The MODEL OUTPUT should be a valid python code."""
    },
)


@base_eval_python_code.wrap
def eval_python_code(
    evaluate: EvaluateCall, row, task_result: patronus.TaskResult, **kwargs
):
    return evaluate(
        evaluated_model_input=row.input,
        evaluated_model_output=task_result.evaluated_model_output,
    )


is_python_code = cli.remote_evaluator(
    "judge",
    criteria="is-pythonic",
    criteria_config={"pass_criteria": """The MODEL OUTPUT should be pythonic code."""},
)


@is_python_code.warp
async def is_idiomatic_python(
    evaluate, row: patronus.Row, parent: patronus.EvalParent, experiment_id
):
    if not parent:
        return None
    if (
        e := parent.find_eval_result(eval_python_code)
    ) is not None and e.pass_ is False:
        return None

    return await evaluate(
        evaluated_model_input=row.evaluated_model_input,
        evaluated_model_output=parent.task.evaluated_model_output,
        experiment_id=experiment_id,
    )


@patronus.task
def make_code_pythonic(
    row, parent: patronus.EvalParent
) -> typing.Optional[patronus.TaskResult]:
    if not parent or not parent.parent:
        return None

    model_output = (
        oai.chat.completions.create(
            model="gpt-4o-mini",
            messages=[
                {
                    "role": "system",
                    "content": "Given the piece of code make sure it's written idiomatically in given language.",
                },
                {"role": "user", "content": parent.parent.task.evaluated_model_output},
            ],
            temperature=0,
        )
        .choices[0]
        .message.content
    )
    return model_output


ex = cli.experiment(
    "protego-dev",
    dataset=dataset,
    chain=[
        {"task": call_llm, "evaluators": [eval_python_code]},
        {"task": ..., "evaluators": [is_idiomatic_python]},
    ],
)

df = ex.to_dataframe()

df.to_csv("./demo.csv")



from patronus import Client

client = Client()

detect_pii = client.remote_evaluator("pii")

client.experiment(
    "Tutorial",
    dataset=[
        {
            "evaluated_model_input": "Please provide your contact details.",
            "evaluated_model_output": "My email is john.doe@example.com and my phone number is 123-456-7890.",
        },
        {
            "evaluated_model_input": "Share your personal information.",
            "evaluated_model_output": "My name is Jane Doe and I live at 123 Elm Street.",
        },
    ],
    evaluators=[detect_pii],
    experiment_name="Detect PII",
)



import patronus


client = patronus.Client()

ex_match = client.remote_evaluator("exact-match", "patronus:exact-match")

dataset = [
        {"evaluated_model_output": "hello", "evaluated_model_gold_answer": "Hello!"},
]


client.experiment(
        "JR",
        dataset=dataset,
        evaluators=[ex_match],
)


from patronus import Client, task

cli = Client()

import logging

formatter = logging.Formatter("[%(levelname)-5s] [%(name)-10s] %(message)s")
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.DEBUG)
console_handler.setFormatter(formatter)

plog = logging.getLogger("patronus")
plog.setLevel(logging.DEBUG)
plog.propagate = False
plog.addHandler(console_handler)

eval_patronus_is_similar = cli.remote_evaluator(
    evaluator_id_or_alias="custom-small",
    criteria="system:is-similar-to-gold-answer",
    max_attempts=4,
)


@task
def say_hi(evaluated_model_input: str) -> str:
    return f"Hi {evaluated_model_input}"


entry = {"evaluated_model_input": "Foo", "evaluated_model_gold_answer": "Hi Foo"}

ex = cli.experiment(
    "patronus-evals",
    data=[entry] * 20,
    task=say_hi,
    evaluators=[eval_patronus_is_similar],
)

ex.to_dataframe()



from enum import Enum


class EventNames(str, Enum):
    evaluation_data = "pat.evaluation.data"


class Attributes(str, Enum):
    event_name = "event.name"
    log_id = "pat.log.id"
    log_type = "pat.log.type"
    span_type = "pat.span.type"
    project_name = "pat.project.name"
    app = "pat.app"
    experiment_id = "pat.experiment.id"
    experiment_name = "pat.experiment.name"
    evaluator_id = "pat.evaluator.id"
    evaluator_criteria = "pat.evaluator.criteria"


class GenAIAttributes(str, Enum):
    operation_name = "gen_ai.operation.name"


class OperationNames(str, Enum):
    eval = "evaluation"
    task = "task"


class LogTypes(str, Enum):
    # Eval log type is emitted by evaluators, they contain evaluation data
    eval = "eval"
    # Trace log type is emitted by traced() decorator.
    trace = "trace"
    # User log type is emitted any time user uses the logger directly.
    user = "user"


class SpanTypes(str, Enum):
    eval = "eval"
    experiment_sample = "experiment.sample"
    experiment_chain_step = "experiment.chain.step"
    experiment_task = "experiment.task"



import collections

import functools

import inspect


def arg_logger(fn):
    sig = inspect.signature(fn)
    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        print(f"AM I Method? {inspect.ismethodwrapper(fn)}")
        print(sig.parameters)
        b_args = sig.bind(*args, **kwargs)
        print(isinstance(b_args.arguments, collections.OrderedDict))
        print(b_args.arguments)
        b_args.arguments.popitem()
        print(b_args.arguments)

    return wrapper

class Bar:
    @arg_logger
    def foo(self, x, y, *, foo=None, bar=None):
        ...

Bar().foo(1, 2, foo="foo", bar="bar")


import random
from typing import Optional

import patronus

client = patronus.Client()


@client.register_local_evaluator("reverse")
def my_local_evaluator(
    evaluated_model_system_prompt: Optional[str],
    # evaluated_model_retrieved_context: Optional[Union[list[str], str]],
    evaluated_model_input: Optional[str],
    evaluated_model_output: Optional[str],
    # evaluated_model_gold_answer: Optional[str],
    # evaluated_model_attachments: Optional[list[dict[str, Any]]],
    # tags: Optional[dict[str, str]] = None,
    # explain_strategy: Literal["never", "on-fail", "on-success", "always"] = "always",
    **kwargs,
) -> patronus.EvaluationResult:
    v = random.random()
    pass_ = v < 0.66
    if pass_ < 0.33:
        pass_ = None
    return patronus.EvaluationResult(
        pass_=pass_,
        score_raw=v,
        text_output=evaluated_model_output[::-1],
        metadata={
            "system_prompt": evaluated_model_system_prompt
            and evaluated_model_system_prompt[::-1],
            "input": evaluated_model_input and evaluated_model_input[::-1],
            "output": evaluated_model_output and evaluated_model_output[::-1],
        },
        explanation="An explanation!",
        evaluation_duration_s=random.random(),
        explanation_duration_s=random.random(),
        tags={"env": "local"},
    )


resp = client.evaluate(
    "reverse",
    evaluated_model_system_prompt="sys prompt!",
    evaluated_model_input="Say Foo",
    evaluated_model_output="Foo!",
    evaluated_model_attachments=[
        {"url": "http://localhost:1222/foo/bar", "media_type": "image/jpeg"}
    ],
)
print(resp.model_dump(by_alias=True))



import asyncio

from patronus import Client, evaluator, Row

client = Client()


@evaluator
async def iexact_match(row: Row) -> bool:
    print(f"[{row.sid}] started...")
    await asyncio.sleep(1)
    return (
        row.evaluated_model_output.lower().strip()
        == row.evaluated_model_gold_answer.lower().strip()
    )


client.experiment(
    "Tutorial",
    dataset=[
        {
            "evaluated_model_input": "Translate 'Good night' to French.",
            "evaluated_model_output": "bonne nuit",
            "evaluated_model_gold_answer": "Bonne nuit",
        },
        {
            "evaluated_model_input": "Summarize: 'AI improves efficiency'.",
            "evaluated_model_output": "ai improves efficiency",
            "evaluated_model_gold_answer": "AI improves efficiency",
        },
    ]
    * 20,
    evaluators=[iexact_match],
    experiment_name="Case Insensitive Match",
    max_concurrency=3,
)



import typing

from typing import Optional, Union, Any

import patronus
from patronus import StructuredEvaluator, EvaluationResult
from patronus.api.api_types import Evaluation
from patronus.evals import RemoteEvaluator, evaluator, bundled_eval
from opentelemetry.instrumentation.httpx import HTTPXClientInstrumentor


patronus.init(integrations=HTTPXClientInstrumentor())


@evaluator()
def is_even(n):
    return n % 2 == 0


class ExactMatch(StructuredEvaluator):

    def evaluate(self, *, system_prompt: Optional[str] = None, task_context: Union[list[str], str, None] = None,
                 task_attachments: Union[list[Any], None] = None, task_input: Optional[str] = None,
                 task_output: Optional[str] = None, gold_answer: Optional[str] = None,
                 task_metadata: Optional[typing.Dict[str, typing.Any]] = None, **kwargs: Any) -> EvaluationResult:
        return EvaluationResult(pass_=task_output==gold_answer)


@patronus.traced("workflow")
def main():
    with patronus.Patronus() as c:
        c.evaluate(
            evaluators=[("judge", "patronus:is-polite"), ExactMatch()],
            task_output="Hello! :) How can I assist you today?"
        )
    RemoteEvaluator("judge", "patronus:is-polite").evaluate(task_output="Hello?")


main()

# RemoteEvaluator("judge", "patronus:is-polite").evaluate(task_output="Hello?")


import typing
import abc

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from patronus import context


class BasePatronusIntegrator(abc.ABC):
    """
    Abstract base class for Patronus integrations.

    This class defines the interface for integrating external libraries and tools
    with the Patronus context. All specific integrators should inherit from this
    class and implement the required methods.
    """

    @abc.abstractmethod
    def apply(self, ctx: "context.PatronusContext", **kwargs: typing.Any):
        """
        Apply the integration to the given Patronus context.

        This method must be implemented by subclasses to define how the
        integration is applied to a Patronus context instance.

        Args:
            ctx: The Patronus context to apply the integration to.
            **kwargs: Additional keyword arguments specific to the implementation.
        """



import os

from opentelemetry import trace
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import (
    BatchSpanProcessor,
)

from patronus import Client

# Initialize Patronus Opentelemetry
trace_provider = TracerProvider()
trace_processor = BatchSpanProcessor(
    OTLPSpanExporter(
        endpoint=os.environ.get("OTEL_EXPORTER_OTLP_ENDPOINT"),
        headers=os.environ.get("OTEL_EXPORTER_OTLP_HEADERS"),
    )
)
trace_provider.add_span_processor(trace_processor)
trace.set_tracer_provider(trace_provider)

# Get new tracer
tracer = trace.get_tracer("Demo Tracing")

# Initialize Patronus Client
client = Client(api_key=os.environ.get("PATRONUS_API_KEY"))

# Start Opentelemetry tracing
with tracer.start_as_current_span("span-name") as span:
    # Run evaluation
    result = client.evaluate(
        evaluator="lynx-small",
        criteria="patronus:hallucination",
        evaluated_model_input="What is the largest animal in the world?",
        evaluated_model_output="The giant sandworm.",
        evaluated_model_retrieved_context="The blue whale is the largest known animal.",
        tags={"scenario": "onboarding"},
    )



import logging

import asyncio
import random
import threading
import time
import typing
from opentelemetry.trace import get_current_span
from typing import Optional, Union

import patronus
from patronus import evals, traced, EvaluationResult, StructuredEvaluator, AsyncStructuredEvaluator
from patronus.evals import bundled_eval

patronus.init()
log = patronus.get_logger()
log.addHandler(logging.StreamHandler())


EVALUATOR = "judge"
CRITERIA = "patronus:fuzzy-match"
# EVALUATOR = "dummy-local-1"
# CRITERIA = "patronus:fifty-fifty"


@evals.evaluator()
def scorer_a(x: int, y: int):
    span = get_current_span()
    print(f"scorer a, current span: {span}")
    time.sleep(random.random() / 10)
    return x == y


@evals.evaluator()
async def scorer_b(x: int, y: int):
    await asyncio.sleep(random.random() / 10)
    return x**2 == y


class ClassScorer(evals.Evaluator):

    def evaluate(self, x: int, y: int) -> Optional[EvaluationResult]:
        time.sleep(random.random() / 10)
        return EvaluationResult(pass_=x==y)


@patronus.traced()
async def main():
    x, y = 2, 4

    class_scorer = ClassScorer()

    with evals.evaluators.bundled_eval():
        await asyncio.gather(
            asyncio.to_thread(scorer_a, x, y),
            scorer_b(x, y),
            asyncio.to_thread(class_scorer.evaluate, x, y),
        )


@patronus.traced()
def threaded_workflow():
    x, y = 2, 4
    class_scorer = ClassScorer()

    with evals.evaluators.bundled_eval():
        t1 = threading.Thread(target=scorer_a, args=(x, y))
        t2 = threading.Thread(target=scorer_a, args=(x, y))
        t3 = threading.Thread(target=class_scorer.evaluate, args=(x, y))
        t1.start(), t2.start(), t3.start()
        t1.join(), t2.join(), t3.join()

    print("Done.")


class IExactMatch(StructuredEvaluator):
    def evaluate(
        self,
        *,
        system_prompt: Optional[str] = None,
        task_context: Union[list[str], str, None] = None,
        task_input: Optional[str] = None,
        task_output: Optional[str] = None,
        gold_answer: Optional[str] = None,
        task_metadata: Optional[typing.Dict[str, typing.Any]] = None,
        **kwargs: typing.Any,
    ) -> EvaluationResult:
        time.sleep(random.random())
        try:
            return EvaluationResult(pass_=task_output.lower().strip() == gold_answer.lower().strip())
        finally:
            print("IExactMatch finished.")


class ExactMatch(StructuredEvaluator):

    def evaluate(
        self,
        *,
        system_prompt: Optional[str] = None,
        task_context: Union[list[str], str, None] = None,
        task_input: Optional[str] = None,
        task_output: Optional[str] = None,
        gold_answer: Optional[str] = None,
        task_metadata: Optional[typing.Dict[str, typing.Any]] = None,
        **kwargs: typing.Any,
    ) -> EvaluationResult:
        time.sleep(random.random())
        try:
            return EvaluationResult(pass_=task_output == gold_answer)
        finally:
            print("ExactMatch finished.")


class AsyncExactMatch(AsyncStructuredEvaluator):

    async def evaluate(
        self,
        *,
        system_prompt: Optional[str] = None,
        task_context: Union[list[str], str, None] = None,
        task_input: Optional[str] = None,
        task_output: Optional[str] = None,
        gold_answer: Optional[str] = None,
        task_metadata: Optional[typing.Dict[str, typing.Any]] = None,
        **kwargs: typing.Any,
    ) -> EvaluationResult:
        await asyncio.sleep(random.random())
        try:
            return EvaluationResult(pass_=task_output == gold_answer)
        finally:
            print(f"{self.__class__.__qualname__} finished.")

iexact_match = IExactMatch()
exact_match = ExactMatch()
async_exact_match = AsyncExactMatch()


@traced()
def raiser():
    time.sleep(random.random() / 10)
    raise Exception("Hello exception")


@traced()
def oneoff():
    score = scorer_a(1, 2)
    print(score)
    try:
        raiser()
    except Exception:
        pass

@patronus.traced()
def pat_cli_workflow():
    pc = patronus.pat_client.Patronus(workers=4)
    with bundled_eval():
        iexact_match.evaluate(task_output="hello world\t", gold_answer="Hello World")
        exact_match.evaluate(task_output="hello world\t", gold_answer="Hello World")
    res = pc.evaluate(
        evaluators=[iexact_match, exact_match],
        task_output="hello world\t",
        gold_answer="Hello World",
    )
    print(res)

@traced()
def workflow_in_bg():
    with patronus.pat_client.Patronus(shutdown_on_exit=False) as pc:
        resp = pc.evaluate_bg(
            evaluators=[iexact_match, exact_match],
            task_output="hello world\t",
            gold_answer="Hello World",
        )
        evals = resp.get()
        evals.raise_on_exception()
        print(f"all succeeded: {evals.all_succeeded()}")
        print(f"any failed: {evals.any_failed()}")
        print(f"Succeeded evals: {list(evals.succeeded_evaluations())}")
        print(f"Failed evals: {list(evals.failed_evaluations())}")

@traced()
async def workflow_async_cli():
    patronus.get_logger().info("Staring workflow...")
    await async_exact_match.evaluate(task_output="hello world\t", gold_answer="Hello World")
    async with patronus.pat_client.AsyncPatronus(2) as pc:
        resp = await pc.evaluate(
            evaluators=[async_exact_match, exact_match, iexact_match, async_exact_match, iexact_match],
            task_output="hello world\t",
            gold_answer="Hello World",
        )
    print(resp)

@traced()
async def workflow_async_cli_bg():
    pc = patronus.pat_client.AsyncPatronus()
    resp = pc.evaluate_bg(
        evaluators=[async_exact_match, exact_match, iexact_match],
        task_output="hello world\t",
        gold_answer="Hello World",
    )
    await pc.close()


@traced()
def workflow_remote_eval():
    ev = evals.evaluators.RemoteEvaluator(EVALUATOR, CRITERIA)
    resp = ev.evaluate(task_output="hello world\t", gold_answer="Hello World")
    with patronus.pat_client.Patronus(2) as pc:
        pc.evaluate_bg(
            evaluators=[ev, ev, ev],
            task_output="hello world\t",
            gold_answer="Hello World"
        )
    print("Response")
    print("========")
    print(resp)

@traced()
async def workflow_remote_eval_async():
    ev = evals.evaluators.AsyncRemoteEvaluator(EVALUATOR, CRITERIA)
    # resp = await ev.evaluate(task_output="hello world\t", gold_answer="Hello World")
    async with patronus.pat_client.AsyncPatronus(2) as pc:
        pc.evaluate_bg(
            evaluators=[ev, ev, ev, iexact_match],
            task_output="hello world\t",
            gold_answer="Hello World"
        )
    print("Response")
    print("========")
    # print(resp)


@traced()
async def complete_anthropic():
    await asyncio.sleep(random.random())
    await asyncio.sleep(random.random())
    await asyncio.sleep(random.random())
    log.info("complete_anthropic finished.")


@traced()
async def complete():
    await complete_anthropic()
    await asyncio.sleep(random.random())
    log.info("complete finished.")
    scorer_a(1, 2)

def in_worker():
    asyncio.run(complete())

@traced()
async def agent_workflow():
    await complete()
    await asyncio.sleep(random.random())
    log.info("agent_workflow finished.")


if __name__ == '__main__':
    oneoff()
    asyncio.run(main())
    threaded_workflow()
    pat_cli_workflow()
    workflow_in_bg()
    asyncio.run(workflow_async_cli())
    asyncio.run(workflow_async_cli_bg())
    workflow_remote_eval()
    asyncio.run(workflow_remote_eval_async())
    asyncio.run(agent_workflow())



import dataclasses
from typing import Union, Generator, IO, Optional
from io import StringIO

from patronus.evals import EvaluationResult
from patronus.exceptions import MultiException


@dataclasses.dataclass
class EvaluationContainer:
    results: list[Union[EvaluationResult, None, Exception]]

    def format(self) -> str:
        """
        Format the evaluation results into a readable summary.
        """
        buf = StringIO()

        total = len(self.results)
        exceptions_count = sum(1 for r in self.results if isinstance(r, Exception))
        successes_count = sum(1 for r in self.results if isinstance(r, EvaluationResult) and r.pass_ is True)
        failures_count = sum(1 for r in self.results if isinstance(r, EvaluationResult) and r.pass_ is False)

        buf.write(f"Total evaluations: {total}\n")
        buf.write(f"Successes: {successes_count}\n")
        buf.write(f"Failures: {failures_count}\n")
        buf.write(f"Exceptions: {exceptions_count}\n\n")
        buf.write("Evaluation Details:\n")
        buf.write("---\n")

        # Add detailed evaluation results
        for result in self.results:
            if result is None:
                buf.write("None\n")
            elif isinstance(result, Exception):
                buf.write(str(result))
                buf.write("\n")
            else:
                buf.write(result.format())
            buf.write("---\n")

        return buf.getvalue()

    def pretty_print(self, file: Optional[IO] = None) -> None:
        """
        Formats and prints the current object in a human-readable form.

        Args:
            file:
        """
        f = self.format()
        print(f, file=file)

    def has_exception(self) -> bool:
        """
        Checks if the results contain any exception.
        """
        return any(isinstance(r, Exception) for r in self.results)

    def raise_on_exception(self) -> None:
        """
        Checks the results for any exceptions and raises them accordingly.
        """
        if not self.has_exception():
            return None
        exceptions = list(r for r in self.results if isinstance(r, Exception))
        if len(exceptions) == 1:
            raise exceptions[0]
        raise MultiException(exceptions)

    def all_succeeded(self, ignore_exceptions: bool = False) -> bool:
        """
        Check if all evaluations that were actually evaluated passed.

        Evaluations are only considered if they:
        - Have a non-None pass_ flag set
        - Are not None (skipped)
        - Are not exceptions (unless ignore_exceptions=True)

        Note: Returns True if no evaluations met the above criteria (empty case).
        """
        for r in self.results:
            if isinstance(r, Exception) and not ignore_exceptions:
                self.raise_on_exception()
            if r is not None and r.pass_ is False:
                return False
        return True

    def any_failed(self, ignore_exceptions: bool = False) -> bool:
        """
        Check if any evaluation that was actually evaluated failed.

        Evaluations are only considered if they:
        - Have a non-None pass_ flag set
        - Are not None (skipped)
        - Are not exceptions (unless ignore_exceptions=True)

        Note: Returns False if no evaluations met the above criteria (empty case).
        """
        for r in self.results:
            if isinstance(r, Exception) and not ignore_exceptions:
                self.raise_on_exception()
            if r is not None and r.pass_ is False:
                return True
        return False

    def failed_evaluations(self) -> Generator[EvaluationResult, None, None]:
        """
        Generates all failed evaluations from the results.
        """
        return (r for r in self.results if not isinstance(r, (Exception, type(None))) and r.pass_ is False)

    def succeeded_evaluations(self) -> Generator[EvaluationResult, None, None]:
        """
        Generates all successfully passed evaluations from the `results` attribute.
        """
        return (r for r in self.results if not isinstance(r, (Exception, type(None))) and r.pass_ is True)



import typing
from typing import Optional

from patronus import datasets
from patronus.evals import Evaluator, EvaluationResult
from patronus.experiments import run_experiment
from patronus.experiments.adapters import EvaluatorAdapter
from patronus.experiments.types import TaskResult, EvalParent


class MatchEvaluator(Evaluator):
    def __init__(self, sanitizer=None):
        if sanitizer is None:
            sanitizer = lambda x: x
        self.sanitizer = sanitizer

    def evaluate(self, actual: str, expected: str) -> EvaluationResult:
        matched = self.sanitizer(actual) == self.sanitizer(expected)
        return EvaluationResult(pass_=matched, score=int(matched))


exact_match = MatchEvaluator()
fuzzy_match = MatchEvaluator(lambda x: x.strip().lower())


class MatchAdapter(EvaluatorAdapter):
    def __init__(self, evaluator: MatchEvaluator):
        super().__init__(evaluator)

    def transform(
        self,
        row: datasets.Row,
        task_result: Optional[TaskResult],
        parent: EvalParent,
        **kwargs
    ) -> tuple[list[typing.Any], dict[str, typing.Any]]:
        args = [row.task_output, row.gold_answer]
        kwargs = {}
        # Passing arguments via kwargs would also work in this case.
        # kwargs = {"actual": row.task_output, "expected": row.gold_answer}
        return args, kwargs


run_experiment(
    dataset=[{"task_output": "string\t", "gold_answer": "string"}],
    evaluators=[MatchAdapter(exact_match), MatchAdapter(fuzzy_match)],
)



import threading

import random

import time
from opentelemetry.trace import get_current_span, SpanContext

import patronus
from patronus import traced

result: SpanContext = None

@traced()
def run_in_thread():
    global result
    ctx = get_current_span()
    time.sleep(1)
    result = ctx.get_span_context()
    return None


@patronus.traced()
def raiser():
    time.sleep(random.random() / 20)
    span = get_current_span()

    t = threading.Thread(target=run_in_thread)
    t.start()
    t.join()

    span.add_link(result, {"key": "value"})

    span.add_event("We crash soon", {"foo": "bar"})
    raise RuntimeError("Hello")

@patronus.traced()
def workflow():
    time.sleep(random.random() / 20)
    patronus.get_logger().info("Hello!")
    raiser()


def main():
    patronus.init()
    workflow()


if __name__ == '__main__':
    main()



from Levenshtein import distance

from patronus import Client, Evaluator, EvaluationResult, simple_task


cli = Client()


class LevenshteinScorer(Evaluator):
    def __init__(self, pass_threshold: float):
        self.pass_threshold = pass_threshold
        super().__init__()

    def evaluate(
        self, evaluated_model_output: str, evaluated_model_gold_answer: str
    ) -> EvaluationResult:
        max_len = max(
            len(x) for x in [evaluated_model_output, evaluated_model_gold_answer]
        )
        score = 1
        if max_len > 0:
            score = 1 - (
                distance(evaluated_model_output, evaluated_model_gold_answer) / max_len
            )

        return EvaluationResult(
            score_raw=score,
            pass_=score >= self.pass_threshold,
            tags={"pass_threshold": str(self.pass_threshold)},
        )


cli.experiment(
    "LevenshteinScorer",
    data=[
        {
            "evaluated_model_input": "Foo",
            "evaluated_model_gold_answer": "Hi Foo",
        },
        {
            "evaluated_model_input": "Bar",
            "evaluated_model_gold_answer": "Hello Bar",
        },
        {
            "evaluated_model_input": "Bar",
            "evaluated_model_gold_answer": "eloh Bar!",
        },
        {
            "evaluated_model_input": "Hello Foo, how have you been?",
            "evaluated_model_gold_answer": "Hi Foo, nice to see you.",
        },
        {
            "evaluated_model_input": "Hi Bar, it's been a long time since we last talked.",
            "evaluated_model_gold_answer": "Hey Bar, how are you?",
        },
        {
            "evaluated_model_input": "Good day Baz, what have you been up to lately?",
            "evaluated_model_gold_answer": "Hi Baz, how have you been?",
        },
        {
            "evaluated_model_input": "Hello Qux, I hope you are doing well.",
            "evaluated_model_gold_answer": "Hello Qux, great to meet you.",
        },
        {
            "evaluated_model_input": "Hi Quux, how is everything going on your side?",
            "evaluated_model_gold_answer": "Hi Quux, hope you're well.",
        },
        {
            "evaluated_model_input": "Hey Corge, it's nice to catch up with you after so long.",
            "evaluated_model_gold_answer": "Hello Corge, what's new?",
        },
        {
            "evaluated_model_input": "Hi Grault, I hope you had a good weekend.",
            "evaluated_model_gold_answer": "Hello Grault, are you ready for the meeting?",
        },
        {
            "evaluated_model_input": "Hello Garply, how have you been these days?",
            "evaluated_model_gold_answer": "Hi Garply, any plans for the weekend?",
        },
        {
            "evaluated_model_input": "Hi Waldo, it's been great working with you.",
            "evaluated_model_gold_answer": "Hello Waldo, hope your day is going great.",
        },
        {
            "evaluated_model_input": "Hello Fred, how did the project go?",
            "evaluated_model_gold_answer": "Hi Fred, did you finish the project?",
        },
        {
            "evaluated_model_input": "Hey Plugh, what's new in your world?",
            "evaluated_model_gold_answer": "Hello Plugh, what are you working on today?",
        },
        {
            "evaluated_model_input": "Hi Xyzzy, welcome to the team.",
            "evaluated_model_gold_answer": "Hi Xyzzy, welcome to the new office!",
        },
        {
            "evaluated_model_input": "Hello Thud, how's everything on your side?",
            "evaluated_model_gold_answer": "Hello Thud, how are things on your end?",
        },
        {
            "evaluated_model_input": "Hi Zelda, great to have you with us.",
            "evaluated_model_gold_answer": "Hi Zelda, good to have you with us.",
        },
    ],
    task=simple_task(lambda input: f"Hi {input}"),
    evaluators=[LevenshteinScorer(pass_threshold=0.6)],
)



from typing import Optional


class WithStackTraceMixin:
    stack_trace: Optional[str] = None


class MultiException(Exception):
    def __init__(self, exceptions: list[Exception]):
        self.exceptions = exceptions
        super().__init__(f"Multiple exception occurred: {self.exceptions}")


class UninitializedError(Exception): ...



import logging

import patronus


formatter = logging.Formatter("[%(levelname)-5s] [%(name)-10s] %(message)s")
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.DEBUG)
console_handler.setFormatter(formatter)

plog = logging.getLogger("patronus")
plog.setLevel(logging.DEBUG)
plog.propagate = False
plog.addHandler(console_handler)


dataset = patronus.read_csv("./out.csv")

cli = patronus.Client()

dummy_eval = cli.remote_evaluator("dummy")

cli.experiment(
    "protego-dev",
    dataset=dataset,
    evaluators=[dummy_eval],
)



import dataclasses
import functools
import logging
import typing
import uuid
from time import time_ns
from types import MappingProxyType
from typing import Optional, Union

from opentelemetry._logs import SeverityNumber
from opentelemetry.exporter.otlp.proto.grpc._log_exporter import OTLPLogExporter as OTLPLogExporterTCP
from opentelemetry.sdk._logs import LogRecord
from opentelemetry.sdk._logs import Logger as OTELLogger
from opentelemetry.sdk._logs import LoggerProvider as OTELLoggerProvider
from opentelemetry.sdk._logs import LoggingHandler as OTeLLoggingHandler
from opentelemetry.sdk._logs._internal import ConcurrentMultiLogRecordProcessor, SynchronousMultiLogRecordProcessor
from opentelemetry.sdk._logs.export import BatchLogRecordProcessor
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.util.instrumentation import InstrumentationScope
from opentelemetry.trace import get_current_span, SpanContext
from opentelemetry.util.types import Attributes as OTeLAttributes

from patronus import context
from patronus.context.context_utils import ResourceMutex
from patronus.tracing.attributes import Attributes, LogTypes


@dataclasses.dataclass
class PatronusScope:
    project_name: Optional[str]
    app: Optional[str]
    experiment_id: Optional[str]
    experiment_name: Optional[str]

    @functools.cached_property
    def as_attributes(self) -> dict[str, str]:
        attrs = {}
        if self.project_name:
            attrs[Attributes.project_name.value] = self.project_name
        if self.app:
            attrs[Attributes.app.value] = self.app
        if self.experiment_id:
            attrs[Attributes.experiment_id.value] = self.experiment_id
        if self.experiment_name:
            attrs[Attributes.experiment_name.value] = self.experiment_name
        return attrs


class LoggerProvider(OTELLoggerProvider):
    project_name: Optional[str]
    app: Optional[str]
    experiment_id: Optional[str]
    experiment_name: Optional[str]

    def __init__(
        self,
        service: Optional[str] = None,
        project_name: Optional[str] = None,
        app: Optional[str] = None,
        experiment_id: Optional[str] = None,
        experiment_name: Optional[str] = None,
        shutdown_on_exit: bool = True,
        multi_log_record_processor: Union[
            SynchronousMultiLogRecordProcessor,
            ConcurrentMultiLogRecordProcessor,
        ] = None,
    ):
        self.project_name = project_name
        self.app = app
        self.experiment_id = experiment_id
        self.experiment_name = experiment_name

        resource = None
        if service is not None:
            resource = Resource.create({"service.name": service})
        super().__init__(resource, shutdown_on_exit, multi_log_record_processor)

    def _get_logger_no_cache(
        self,
        name: str,
        version: Optional[str] = None,
        schema_url: Optional[str] = None,
        attributes: Optional[OTeLAttributes] = None,
    ) -> "Logger":
        pat_scope = PatronusScope(
            project_name=self.project_name,
            app=self.app,
            experiment_id=self.experiment_id,
            experiment_name=self.experiment_name,
        )
        attributes = attributes or {}
        attributes = {**attributes, **pat_scope.as_attributes}
        return Logger(
            self._resource,
            self._multi_log_record_processor,
            InstrumentationScope(
                name,
                version,
                schema_url,
                attributes,
            ),
            pat_scope,
        )

    def get_logger(self, *args, **kwargs) -> "Logger":
        return super().get_logger(*args, **kwargs)


class LoggingHandler(OTeLLoggingHandler):
    def __init__(self, pat_scope: PatronusScope, level=logging.NOTSET, logger_provider=None) -> None:
        super().__init__(level, logger_provider)
        self.pat_scope = pat_scope

    def emit(self, record: logging.LogRecord) -> None:
        # Add patronus attributes to all outgoing logs.
        # This handler transforms record into OTeL LogRecord.
        for k, v in self.pat_scope.as_attributes.items():
            # We use setattr since attributes may have dots in their names
            # and record is an object, not a dict.
            setattr(record, k, v)
        setattr(record, Attributes.log_type, LogTypes.user.value)
        super().emit(record)


def encode_attrs(v):
    if isinstance(v, dict):
        keys = list(v.keys())
        for k in keys:
            v[k] = encode_attrs(v[k])
        return v
    if isinstance(v, (list, tuple, str, int, float, bool, type(None))):
        return v
    return str(v)


def transform_body(v: typing.Any):
    if v is None:
        return None
    if isinstance(v, MappingProxyType):
        v = dict(v)
    if isinstance(v, list):
        return [transform_body(vv) for vv in v]
    if isinstance(v, tuple):
        return tuple(transform_body(vv) for vv in v)
    if isinstance(v, (str, bool, int, float)):
        return v
    if not isinstance(v, dict):
        return str(v)

    return {str(k): transform_body(v) for k, v in v.items()}


class Logger(OTELLogger):
    def __init__(
        self,
        resource: Resource,
        multi_log_record_processor: Union[
            SynchronousMultiLogRecordProcessor,
            ConcurrentMultiLogRecordProcessor,
        ],
        instrumentation_scope: InstrumentationScope,
        pat_scope: PatronusScope,
    ):
        super().__init__(resource, multi_log_record_processor, instrumentation_scope)
        self.pat_scope = pat_scope

    def log(
        self,
        body: typing.Any,
        *,
        log_attrs: Optional[OTeLAttributes] = None,
        severity: Optional[SeverityNumber] = None,
        log_type: LogTypes = LogTypes.user,
        log_id: Optional[uuid.UUID] = None,
        span_context: Optional[SpanContext] = None,
        event_name: Optional[str] = None,
    ) -> uuid.UUID:
        severity: SeverityNumber = severity or SeverityNumber.INFO
        span_context = span_context or get_current_span().get_span_context()

        log_id = log_id or uuid.uuid4()
        log_attrs = log_attrs or {}
        log_attrs.update(
            {
                Attributes.log_id.value: str(log_id),
                Attributes.log_type.value: log_type.value,
            }
        )
        if event_name is not None:
            log_attrs[Attributes.event_name.value] = event_name
        log_attrs.update(self.pat_scope.as_attributes)

        record = LogRecord(
            timestamp=time_ns(),
            observed_timestamp=time_ns(),
            trace_id=span_context.trace_id,
            span_id=span_context.span_id,
            trace_flags=span_context.trace_flags,
            severity_text=severity.name,
            severity_number=severity,
            body=transform_body(body),
            attributes=log_attrs,
            resource=self.resource,
        )
        self.emit(record)
        return log_id

    def evaluation_data(
        self,
        *,
        system_prompt: Optional[str] = None,
        task_context: Union[list[str], str, None] = None,
        task_input: Optional[str] = None,
        task_output: Optional[str] = None,
        gold_answer: Optional[str] = None,
        task_metadata: Optional[dict[str, typing.Any]] = None,
        log_attrs: Optional[OTeLAttributes] = None,
    ) -> uuid.UUID:
        if isinstance(task_context, str):
            task_context = [task_context]
        log_attrs = log_attrs or {}
        log_attrs[Attributes.log_type] = LogTypes.eval

        return self.log(
            {
                "system_prompt": system_prompt,
                "task_context": task_context,
                "task_input": task_input,
                "task_output": task_output,
                "gold_answer": gold_answer,
                "task_metadata": task_metadata,
            }
        )


def _create_exporter(endpoint: str, api_key: str) -> OTLPLogExporterTCP:
    return OTLPLogExporterTCP(endpoint=endpoint, headers={"x-api-key": api_key}, insecure=True)


@functools.lru_cache()
def create_logger_provider(exporter_endpoint: str, api_key: str, scope: context.PatronusScope) -> LoggerProvider:
    logger_provider = LoggerProvider(project_name=scope.project_name, app=scope.app, experiment_id=scope.experiment_id)
    exporter = _create_exporter(exporter_endpoint, api_key)
    logger_provider.add_log_record_processor(BatchLogRecordProcessor(exporter))
    return logger_provider


def create_patronus_logger(
    scope: context.PatronusScope,
    exporter_endpoint: str,
    api_key: str,
) -> Logger:
    provider = create_logger_provider(
        exporter_endpoint=exporter_endpoint,
        api_key=api_key,
        scope=scope,
    )
    return provider.get_logger("patronus.sdk")


__logger_count = ResourceMutex(0)


def set_logger_handler(logger: logging.Logger, scope: context.PatronusScope, provider: LoggerProvider):
    if any(isinstance(hdl, LoggingHandler) for hdl in logger.handlers):
        return
    scp = PatronusScope(
        project_name=scope.project_name,
        app=scope.app,
        experiment_id=scope.experiment_id,
        experiment_name=scope.experiment_name,
    )
    logger.addHandler(LoggingHandler(pat_scope=scp, level=logging.NOTSET, logger_provider=provider))


@functools.lru_cache()
def create_logger(
    scope: context.PatronusScope,
    exporter_endpoint: str,
    api_key: str,
) -> logging.Logger:
    provider = create_logger_provider(
        exporter_endpoint=exporter_endpoint,
        api_key=api_key,
        scope=scope,
    )
    with __logger_count as mu:
        n = mu.get()
        mu.set(n + 1)
    if n == 0:
        suffix = ""
    else:
        suffix = f".{n}"
    logger = logging.getLogger(f"patronus.sdk{suffix}")
    pat_scope = PatronusScope(
        project_name=scope.project_name,
        app=scope.app,
        experiment_id=scope.experiment_id,
        experiment_name=scope.experiment_name,
    )
    logger.addHandler(LoggingHandler(pat_scope=pat_scope, level=logging.NOTSET, logger_provider=provider))
    logger.setLevel(logging.DEBUG)
    return logger



import textwrap

from patronus import Client

import logging

formatter = logging.Formatter("[%(levelname)-5s] [%(name)-10s] %(message)s")
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.DEBUG)
console_handler.setFormatter(formatter)

plog = logging.getLogger("patronus")
plog.setLevel(logging.DEBUG)
plog.propagate = False
plog.addHandler(console_handler)

cli = Client()

evaluate_proper_language = cli.remote_evaluator(
    "custom-large",
    "detect-requested-programming-languages",
    criteria_config={
        "pass_criteria": textwrap.dedent(
            """
            The MODEL OUTPUT should provide only valid code in a well-known programming language.
            The MODEL OUTPUT should consist of the code in a programming language specified in the USER INPUT.
            """
        ),
    },
    allow_update=True,
)

data = [
    {
        "evaluated_model_input": "Write a hello world example in Python.",
        "evaluated_model_output": "print('Hello World!')",
    },
    {
        "evaluated_model_input": "Write a hello world example in JavaScript.",
        "evaluated_model_output": "print('Hello World!')",
    },
]

cli.experiment(
    "Programming Language Detection",
    data=data,
    evaluators=[evaluate_proper_language],
)



import atexit
import collections
import itertools
import logging
import os
import threading
import time
import typing

from patronus.api import PatronusAPIClient, api_types
from patronus.utils import Once

logger = logging.getLogger("patronus.core")

_BEE_RESET_ONCE = Once()


# _FlushRequest is based on _FlushRequest from otel sdk trace batch processor/exporter.
#
# Represents a request for the BatchEvaluationExporter to flush evaluations.
class _FlushRequest:
    __slots__ = ["event", "num_evaluations"]

    def __init__(self):
        self.event = threading.Event()
        self.num_evaluations = 0


def backoff(max_value: typing.Optional[int] = None):
    for i in itertools.count(0):
        out = 2**i
        yield min(out, max_value) if max_value else out


# BatchEvaluationExporter is based by otel sdk trace batch processor.
class BatchEvaluationExporter:
    _client: PatronusAPIClient
    done: bool
    schedule_delay: float
    max_queue_size: int
    queue: typing.Deque[api_types.ClientEvaluation]
    condition: threading.Condition
    max_export_batch_size: int
    evaluations_list: list[typing.Optional[api_types.ClientEvaluation]]
    flush_request: typing.Optional[_FlushRequest]
    _pid: int
    _atexit_handler: typing.Optional[typing.Callable[[], None]]

    def __init__(
        self,
        client: PatronusAPIClient,
        *,
        schedule_delay: float = 5.0,
        max_export_batch_size: int = 100,
        max_queue_size: int = 2000,
        shutdown_on_exit: bool = True,
    ):
        self._client = client

        self.done = False
        self.schedule_delay = schedule_delay
        self.max_queue_size = max_queue_size
        self.queue = collections.deque(maxlen=self.max_queue_size)
        self.condition = threading.Condition(threading.Lock())

        self.max_export_batch_size = max_export_batch_size
        # precallocated list to export evaluations
        self.evaluations_list = [None] * self.max_export_batch_size
        self.flush_request = None

        self.worker_thread = threading.Thread(name="PatronusBatchEvaluationExporter", target=self.worker, daemon=True)
        self.worker_thread.start()

        if hasattr(os, "register_at_fork"):
            os.register_at_fork(after_in_child=self._at_fork_reinit)  # pylint: disable=protected-access
        self._pid = os.getpid()

        self._atexit_handler = None
        if shutdown_on_exit:
            self._atexit_handler = atexit.register(self.shutdown)

    def submit(self, data: api_types.ClientEvaluation):
        if self.done:
            logger.warning("Already shutdown, dropping evaluation.")
            return

        if self._pid != os.getpid():
            _BEE_RESET_ONCE.do_once(self._at_fork_reinit)

        if len(self.queue) >= self.max_queue_size:
            logger.info("Queue is full, likely evaluations will be dropped.")

        self.queue.appendleft(data)

        if len(self.queue) >= self.max_export_batch_size:
            with self.condition:
                self.condition.notify()

    def _at_fork_reinit(self):
        self.condition = threading.Condition(threading.Lock())
        self.queue.clear()

        # worker_thread is local to a process, only the thread that issued fork continues
        # to exist. A new worker thread must be started in child process.
        self.worker_thread = threading.Thread(name="PatronusBatchEvaluationExporter", target=self.worker, daemon=True)
        self.worker_thread.start()
        self._pid = os.getpid()

    def worker(self):
        timeout = self.schedule_delay
        flush_request: typing.Optional[_FlushRequest] = None

        while not self.done:
            with self.condition:
                if self.done:
                    # done flag may have changed, avoid waiting
                    break

                flush_request = self._get_and_unset_flush_request()
                if (len(self.queue) < self.max_export_batch_size) and flush_request is None:
                    self.condition.wait(timeout)
                    flush_request = self._get_and_unset_flush_request()

                if not self.queue:
                    timeout = self.schedule_delay
                    self._notify_flush_request_finished(flush_request)
                    flush_request = None
                    continue

                if self.done:
                    break

            start = time.time_ns()
            self._export(flush_request)
            end = time.time_ns()
            elapsed = (end - start) / 1e9
            timeout = self.schedule_delay - elapsed

            self._notify_flush_request_finished(flush_request)
            flush_request = None

        # there might have been a new flush request while export was running
        # and before the done flag switched to true
        with self.condition:
            shutdown_flush_request = self._get_and_unset_flush_request()

        self._drain_queue()
        self._notify_flush_request_finished(flush_request)
        self._notify_flush_request_finished(shutdown_flush_request)

    def _drain_queue(self) -> int:
        exported = 0
        while self.queue:
            exported += self._export_batch()
        return exported

    def _export(self, flush_request: typing.Optional[_FlushRequest] = None) -> None:
        if not flush_request:
            self._export_batch()
            return

        num_evals = flush_request.num_evaluations
        while self.queue:
            num_exported = self._export_batch()
            num_evals -= num_exported

            if num_evals <= 0:
                break

    def _export_batch(self) -> int:
        idx = 0
        while idx < self.max_export_batch_size and self.queue:
            self.evaluations_list[idx] = self.queue.pop()
            idx += 1

        try:
            self.__export_batch(self.evaluations_list[:idx])
        except Exception:
            logger.exception("Failed to export batch, evaluations are dropped most likely.")

        for index in range(idx):
            self.evaluations_list[index] = None

        return idx

    def __export_batch(self, evaluations: typing.List[api_types.ClientEvaluation]) -> None:
        max_delay = 64
        last_exc = None
        for delay in backoff(max_value=max_delay):
            try:
                self._client.batch_create_evaluations_sync(
                    request=api_types.BatchCreateEvaluationsRequest(evaluations=evaluations)
                )
                return
            except Exception as e:
                logger.exception(f"Failed to export evaluations. retrying in {delay} seconds.")
                time.sleep(delay)
                last_exc = e

        if last_exc:
            raise last_exc

    def _get_and_unset_flush_request(self) -> typing.Optional[_FlushRequest]:
        flush_request = self.flush_request
        self.flush_request = None
        if flush_request is not None:
            flush_request.num_evaluations = len(self.queue)
        return flush_request

    @staticmethod
    def _notify_flush_request_finished(flush_request: typing.Optional[_FlushRequest]) -> None:
        if flush_request is not None:
            flush_request.event.set()

    def _get_or_create_flush_request(self) -> _FlushRequest:
        if self.flush_request is None:
            self.flush_request = _FlushRequest()
        return self.flush_request

    def force_flush(self, timeout: typing.Optional[float] = None) -> bool:
        timeout = timeout or self.schedule_delay

        if self.done:
            logger.warning("Already shutdown, ignoring call to force_flush().")
            return True

        with self.condition:
            flush_request = self._get_or_create_flush_request()
            # signal the worker thread to flush and wait for it to finish
            self.condition.notify_all()

        # wait for token to be processed
        ret = flush_request.event.wait(timeout)
        if not ret:
            logger.warning("Timeout was exceeded in force_flush().")
        return ret

    def shutdown(self):
        self._shutdown()
        if self._atexit_handler:
            atexit.unregister(self._atexit_handler)

    def _shutdown(self):
        self.done = True
        with self.condition:
            self.condition.notify_all()
        self.worker_thread.join()



import inspect

import typing

import asyncio

import abc


from typing import Union, Optional

from patronus import evals
from patronus import datasets
from patronus.evals.evaluators import coerce_eval_output_type
from patronus.experiments.types import TaskResult, EvalParent
from patronus.evals import EvaluationResult


class BaseEvaluatorAdapter(abc.ABC):
    """
    Abstract base class for all evaluator adapters.

    Evaluator adapters provide a standardized interface between the experiment framework
    and various types of evaluators (function-based, class-based, etc.).

    All concrete adapter implementations must inherit from this class and implement
    the required abstract methods.
    """

    @property
    @abc.abstractmethod
    def evaluator_id(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def criteria(self) -> Optional[str]:
        pass

    @property
    def canonical_name(self) -> str:
        return f"{self.evaluator_id}:{self.evaluator_id or ''}"

    @abc.abstractmethod
    async def evaluate(
        self,
        row: datasets.Row,
        task_result: Optional[TaskResult],
        parent: EvalParent,
        **kwargs,
    ) -> EvaluationResult: ...


class EvaluatorAdapter(BaseEvaluatorAdapter):
    """
    Adapter for class-based evaluators conforming to the Evaluator or AsyncEvaluator protocol.

    This adapter enables the use of evaluator classes that implement either the Evaluator
    or AsyncEvaluator interface within the experiment framework.

    Attributes:
        evaluator: The evaluator instance to adapt.

    **Examples:**

    ```python
    import typing
    from typing import Optional

    from patronus import datasets
    from patronus.evals import Evaluator, EvaluationResult
    from patronus.experiments import run_experiment
    from patronus.experiments.adapters import EvaluatorAdapter
    from patronus.experiments.types import TaskResult, EvalParent


    class MatchEvaluator(Evaluator):
        def __init__(self, sanitizer=None):
            if sanitizer is None:
                sanitizer = lambda x: x
            self.sanitizer = sanitizer

        def evaluate(self, actual: str, expected: str) -> EvaluationResult:
            matched = self.sanitizer(actual) == self.sanitizer(expected)
            return EvaluationResult(pass_=matched, score=int(matched))


    exact_match = MatchEvaluator()
    fuzzy_match = MatchEvaluator(lambda x: x.strip().lower())


    class MatchAdapter(EvaluatorAdapter):
        def __init__(self, evaluator: MatchEvaluator):
            super().__init__(evaluator)

        def transform(
            self,
            row: datasets.Row,
            task_result: Optional[TaskResult],
            parent: EvalParent,
            **kwargs
        ) -> tuple[list[typing.Any], dict[str, typing.Any]]:
            args = [row.task_output, row.gold_answer]
            kwargs = {}
            # Passing arguments via kwargs would also work in this case.
            # kwargs = {"actual": row.task_output, "expected": row.gold_answer}
            return args, kwargs


    run_experiment(
        dataset=[{"task_output": "string\t", "gold_answer": "string"}],
        evaluators=[MatchAdapter(exact_match), MatchAdapter(fuzzy_match)],
    )
    ```

    """

    evaluator: Union[evals.Evaluator, evals.AsyncEvaluator]

    def __init__(self, evaluator: evals.Evaluator):
        if not isinstance(evaluator, evals.Evaluator):
            raise TypeError(f"{evaluator} is not {evals.Evaluator.__name__}.")
        self.evaluator = evaluator

    @property
    def evaluator_id(self) -> str:
        return self.evaluator.get_evaluator_id()

    @property
    def criteria(self) -> Optional[str]:
        return self.evaluator.get_criteria()

    def transform(
        self,
        row: datasets.Row,
        task_result: Optional[TaskResult],
        parent: EvalParent,
        **kwargs: typing.Any,
    ) -> tuple[list[typing.Any], dict[str, typing.Any]]:
        """
        Transform experiment framework arguments to evaluation method arguments.

        Args:
            row: The data row being evaluated.
            task_result: The result of the task execution, if available.
            parent: The parent evaluation context.
            **kwargs: Additional keyword arguments from the experiment.

        Returns:
            A list of positional arguments to pass to the evaluator function.
            A dictionary of keyword arguments to pass to the evaluator function.
        """

        return (
            [],
            {"row": row, "task_result": task_result, "parent": parent, **kwargs},
        )

    async def evaluate(
        self,
        row: datasets.Row,
        task_result: Optional[TaskResult],
        parent: EvalParent,
        **kwargs: typing.Any,
    ) -> EvaluationResult:
        """
        Evaluate the given row and task result using the adapted evaluator function.

        This method implements the BaseEvaluatorAdapter.evaluate() protocol.

        Args:
            row: The data row being evaluated.
            task_result: The result of the task execution, if available.
            parent: The parent evaluation context.
            **kwargs: Additional keyword arguments from the experiment.

        Returns:
            An EvaluationResult containing the evaluation outcome.
        """
        ev_args, ev_kwargs = self.transform(row, task_result, parent, **kwargs)
        return await self._evaluate(*ev_args, **ev_kwargs)

    async def _evaluate(self, *args: typing.Any, **kwargs: typing.Any) -> TaskResult:
        if isinstance(self.evaluator, evals.AsyncEvaluator):
            return await self.evaluator.evaluate(*args, **kwargs)
        elif isinstance(self.evaluator, evals.Evaluator):
            return await asyncio.to_thread(self.evaluator.evaluate, *args, **kwargs)
        else:
            raise RuntimeError("Invalid evaluator type")


class StructuredEvaluatorAdapter(EvaluatorAdapter):
    """
    Adapter for structured evaluators.
    """

    evaluator: Union[evals.StructuredEvaluator, evals.AsyncStructuredEvaluator]

    def __init__(
        self,
        evaluator: Union[evals.StructuredEvaluator, evals.AsyncStructuredEvaluator],
    ):
        if not isinstance(
            evaluator, (evals.StructuredEvaluator, evals.AsyncStructuredEvaluator)
        ):
            raise TypeError(
                f"{type(evaluator)} is not "
                f"{evals.AsyncStructuredEvaluator.__name__} nor {evals.StructuredEvaluator.__name__}."
            )
        super().__init__(evaluator)

    def transform(
        self,
        row: datasets.Row,
        task_result: Optional[TaskResult],
        parent: EvalParent,
        **kwargs: typing.Any,
    ) -> tuple[list[typing.Any], dict[str, typing.Any]]:
        task_output = row.task_output
        task_metadata = row.task_metadata
        if task_result is not None:
            task_output = task_result.output
            task_metadata = task_result.metadata

        ev_kwargs = dict(
            system_prompt=row.system_prompt,
            task_context=row.task_context,
            task_attachments=row.task_attachments,
            task_input=row.task_input,
            task_output=task_output,
            gold_answer=row.gold_answer,
            task_metadata=task_metadata,
        )
        return [], ev_kwargs


class FuncEvaluatorAdapter(BaseEvaluatorAdapter):
    """
    Adapter class that allows using function-based evaluators with the experiment framework.

    This adapter serves as a bridge between function-based evaluators decorated with `@evaluator()`
    and the experiment framework's evaluation system.
    It handles both synchronous and asynchronous evaluator functions.

    Attributes:
        fn (Callable): The evaluator function to be adapted.

    Notes:
        - The function passed to this adapter must be decorated with `@evaluator()`.
        - The adapter automatically handles the conversion between function results and proper
          evaluation result objects.

    Examples:

        Direct usage with a compatible evaluator function:

        ```python
        from patronus import evaluator
        from patronus.experiments import FuncEvaluatorAdapter, run_experiment
        from patronus.datasets import Row


        @evaluator()
        def exact_match(row: Row, **kwargs):
            return row.task_output == row.gold_answer

        run_experiment(
            dataset=[{"task_output": "string", "gold_answer": "string"}],
            evaluators=[FuncEvaluatorAdapter(exact_match)]
        )
        ```

        Customized usage by overriding the `transform()` method:

        ```python
        from typing import Optional
        import typing

        from patronus import evaluator, datasets
        from patronus.experiments import FuncEvaluatorAdapter, run_experiment
        from patronus.experiments.types import TaskResult, EvalParent


        @evaluator()
        def exact_match(actual, expected):
            return actual == expected


        class AdaptedExactMatch(FuncEvaluatorAdapter):
            def __init__(self):
                super().__init__(exact_match)

            def transform(
                self,
                row: datasets.Row,
                task_result: Optional[TaskResult],
                parent: EvalParent,
                **kwargs
            ) -> tuple[list[typing.Any], dict[str, typing.Any]]:
                args = [row.task_output, row.gold_answer]
                kwargs = {}

                # Alternative: passing arguments via kwargs instead of args
                # args = []
                # kwargs = {"actual": row.task_output, "expected": row.gold_answer}

                return args, kwargs


        run_experiment(
            dataset=[{"task_output": "string", "gold_answer": "string"}],
            evaluators=[AdaptedExactMatch()],
        )
        ```

    """

    def __init__(self, fn: typing.Callable[..., typing.Any]):
        if not hasattr(fn, "_pat_evaluator"):
            raise ValueError(
                f"Passed function {fn.__qualname__} is not an evaluator. "
                "Hint: add @evaluator decorator to the function."
            )
        self.fn = fn

    @property
    def evaluator_id(self) -> str:
        # @evaluator() wrapper sets that value
        return self.fn._pat_evaluator_id  # noqa

    @property
    def criteria(self) -> Optional[str]:
        # @evaluator() wrapper sets that value
        return self.fn._pat_criteria  # noqa

    def transform(
        self,
        row: datasets.Row,
        task_result: Optional[TaskResult],
        parent: EvalParent,
        **kwargs: typing.Any,
    ) -> tuple[list[typing.Any], dict[str, typing.Any]]:
        """
        Transform experiment framework parameters to evaluator function parameters.

        Args:
            row: The data row being evaluated.
            task_result: The result of the task execution, if available.
            parent: The parent evaluation context.
            **kwargs: Additional keyword arguments from the experiment.

        Returns:
            A list of positional arguments to pass to the evaluator function.
            A dictionary of keyword arguments to pass to the evaluator function.
        """

        return (
            [],
            {"row": row, "task_result": task_result, "parent": parent, **kwargs},
        )

    async def evaluate(
        self,
        row: datasets.Row,
        task_result: Optional[TaskResult],
        parent: EvalParent,
        **kwargs: typing.Any,
    ) -> EvaluationResult:
        """
        Evaluate the given row and task result using the adapted evaluator function.

        This method implements the BaseEvaluatorAdapter.evaluate() protocol.

        Args:
            row: The data row being evaluated.
            task_result: The result of the task execution, if available.
            parent: The parent evaluation context.
            **kwargs: Additional keyword arguments from the experiment.

        Returns:
            An EvaluationResult containing the evaluation outcome.
        """
        ev_args, ev_kwargs = self.transform(row, task_result, parent, **kwargs)
        return await self._evaluate(*ev_args, **ev_kwargs)

    async def _evaluate(self, *args, **kwargs) -> TaskResult:
        if inspect.iscoroutinefunction(self.fn):
            result = await self.fn(*args, **kwargs)
        else:
            result = await asyncio.to_thread(self.fn, *args, **kwargs)

        return coerce_eval_output_type(result, self.fn.__qualname__)



import random
import time

import patronus
from patronus import Client, task, Row, TaskResult

client = Client()


@task
def my_task(row: Row):
    time.sleep(random.random())
    if row.sid == 4:
        return 123
    return f"{row.evaluated_model_input} World"


@patronus.evaluator
def exact_match(row: Row, task_result: TaskResult):
    time.sleep(random.random())
    return task_result.evaluated_model_output == row.evaluated_model_gold_answer


ex = client.experiment(
    "Hello World",
    dataset=[
        {
            "evaluated_model_input": "Hello",
            "evaluated_model_retrieved_context": "Hello",
            "evaluated_model_gold_answer": "Hello World",
        },
        {
            "evaluated_model_input": "Hello",
            "evaluated_model_retrieved_context": ["Hello", "World"],
            "evaluated_model_gold_answer": "Hello World",
        },
        {
            "evaluated_model_input": "Hello",
            "evaluated_model_retrieved_context": None,
            "evaluated_model_gold_answer": "Hello World",
        },
        {
            "evaluated_model_input": "Hello",
            "evaluated_model_gold_answer": "Hello World",
        },
    ]
    * 20,
    task=my_task,
    evaluators=[exact_match],
    max_concurrency=3,
)

df = ex.to_dataframe()
# print(df.info())
# print(df)


# ex2 = client.experiment(
#     "Hello World",
#     dataset=df,
#     task=my_task,
#     evaluators=[exact_match],
# )
# print("!!!!!!! EX2 !!!!!!!!!")
# print(ex2.to_dataframe().info())



from .datasets import Attachment as Attachment
from .datasets import Dataset as Dataset
from .datasets import DatasetLoader as DatasetLoader
from .datasets import Fields as Fields
from .datasets import Row as Row
from .datasets import read_csv as read_csv
from .datasets import read_jsonl as read_jsonl
from .remote import RemoteDatasetLoader as RemoteDatasetLoader



import asyncio
import functools
import inspect
import typing
from concurrent.futures import ThreadPoolExecutor

T = typing.TypeVar("T")


async def run_as_coro(
    __loop: asyncio.AbstractEventLoop,
    __executor: ThreadPoolExecutor,
    __fn: typing.Callable[[typing.Any, ...], T],
    *args,
    **kwargs,
) -> typing.Awaitable[T]:
    if inspect.iscoroutinefunction(__fn):
        return await __fn(*args, **kwargs)

    @functools.wraps(__fn)
    def inner(args: tuple[typing.Any, ...], kwargs: dict[str, typing.Any]):
        return __fn(*args, **kwargs)

    return await __loop.run_in_executor(__executor, inner, args, kwargs)


def run_until_complete(coro: typing.Awaitable[T]) -> typing.Awaitable[T]:
    try:
        loop = asyncio.get_running_loop()
        return loop.create_task(coro)
    except RuntimeError:
        # Ignore this exception since we can continue by creating an event loop.
        # We are not creating an event loop in the except block to no include
        # this stacktrace part in case of a raised exception in the coroutine.
        # Otherwise, we would get message as
        #
        # > RuntimeError: no running event loop
        # > During handling of the above exception, another exception occurred: [...]
        #
        # Which we don't want because the try catch is just a check whether there is a running event loop.
        # We are not really handling an "exception" here.
        pass

    loop = asyncio.get_event_loop()
    return loop.run_until_complete(coro)



import asyncio

import patronus
from patronus import EvaluateRequest

cli = patronus.Client()

eval_is_code = cli.remote_evaluator("judge", "patronus:is-code")
await cli.api.evaluate(
    EvaluateRequest(
        evaluators=[{"evaluator": "judge", "profile_name": "patronus:is-code"}],
        evaluated_model_output="print('hello world')",
        app="protego-dev-no-ex",
    )
)

# await eval_is_code.evaluate(
#     evaluated_model_output="""
#         ```python
#         print("Hello World")
#         ```
#         """,
#     app="protego-dev-no-ex",
# )
#


async def main():
    resp = await eval_is_code.evaluate(
        evaluated_model_output="""
        ```python
        print("Hello World")
        ```
        """,
        app="protego-dev-no-ex",
    )
    print(resp)
    print()
    print()

    resp2 = await cli.api.evaluate(
        EvaluateRequest(
            evaluators=[{"evaluator": "judge", "profile_name": "patronus:is-code"}],
            evaluated_model_output="print('hello world')",
            app="protego-dev-no-ex",
        )
    )
    print(resp2)


asyncio.run(main())



from transformers import BertTokenizer, BertModel
import numpy as np
from patronus import Client, Evaluator, EvaluationResult, Row

client = Client()


class BERTScore(Evaluator):
    def __init__(self, pass_threshold: float):
        self.pass_threshold = pass_threshold
        self.tokenizer = BertTokenizer.from_pretrained("bert-base-uncased")
        self.model = BertModel.from_pretrained("bert-base-uncased")

        super().__init__()

    def evaluate(self, row: Row) -> EvaluationResult:
        # Tokenize text
        output_toks = self.tokenizer(row.evaluated_model_output, return_tensors="pt", padding=True, truncation=True)
        gold_answer_toks = self.tokenizer(
            row.evaluated_model_gold_answer, return_tensors="pt", padding=True, truncation=True
        )

        # Obtain embeddings from BERT model
        output_embeds = self.model(**output_toks).last_hidden_state.mean(dim=1).detach().numpy()
        gold_answer_embeds = self.model(**gold_answer_toks).last_hidden_state.mean(dim=1).detach().numpy()

        # Calculate cosine similarity
        score = np.dot(output_embeds, gold_answer_embeds.T) / (
            np.linalg.norm(output_embeds) * np.linalg.norm(gold_answer_embeds)
        )

        return EvaluationResult(
            score_raw=score,
            pass_=score >= self.pass_threshold,
            tags={"pass_threshold": str(self.pass_threshold)},
        )


client.experiment(
    "Tutorial",
    dataset=[
        {
            "evaluated_model_input": "Translate 'Goodbye' to Spanish.",
            "evaluated_model_output": "Hasta luego",
            "evaluated_model_gold_answer": "Adiós",
        },
        {
            "evaluated_model_input": "Summarize: 'The quick brown fox jumps over the lazy dog'.",
            "evaluated_model_output": "Quick brown fox jumps over dog",
            "evaluated_model_gold_answer": "The quick brown fox jumps over the lazy dog",
        },
    ],
    evaluators=[BERTScore(pass_threshold=0.8)],
    experiment_name="BERTScore Output Label Similarity",
)



import os
from patronus import Client, Row, TaskResult, evaluator, task

client = Client(
    # This is the default and can be omitted
    api_key=os.environ.get("PATRONUS_API_KEY"),
)


@task
def my_task(row: Row):
    return f"{row.evaluated_model_input} World"


@evaluator
def exact_match(row: Row, task_result: TaskResult):
    return task_result.evaluated_model_output == row.evaluated_model_gold_answer


client.experiment(
    "Tutorial Project",
    dataset=[
        {
            "evaluated_model_input": "Hello",
            "evaluated_model_gold_answer": "Hello World",
        },
    ],
    task=my_task,
    evaluators=[exact_match],
)



from typing import Optional

import typing

from patronus import evaluator, datasets
from patronus.experiments import FuncEvaluatorAdapter, run_experiment
from patronus.experiments.types import TaskResult, EvalParent


@evaluator()
def exact_match(actual, expected):
    return actual == expected


class AdaptedExactMatch(FuncEvaluatorAdapter):
    def __init__(self):
        super().__init__(exact_match)

    def transform(
        self,
        row: datasets.Row,
        task_result: Optional[TaskResult],
        parent: EvalParent,
        **kwargs
    ) -> tuple[list[typing.Any], dict[str, typing.Any]]:
        args = [row.task_output, row.gold_answer]
        kwargs = {}
        # Passing arguments via kwargs would also work in this case.
        # kwargs = {"actual": row.task_output, "expected": row.gold_answer}
        return args, kwargs


run_experiment(
    dataset=[{"task_output": "string", "gold_answer": "string"}],
    evaluators=[AdaptedExactMatch()],
)



#!/usr/bin/env -S uv run --script
# /// script
# requires-python = ">=3.10"
# dependencies = [
#     "patronus",
#     "anthropic",
#     "openinference-instrumentation-anthropic>=0.1.17,<0.2.0",
#     "opentelemetry-instrumentation-threading>=0.54b0,<1.0",
#     "opentelemetry-instrumentation-asyncio>=0.54b0,<1.0"
# ]
# ///
import json
import requests

import anthropic
from openinference.instrumentation.anthropic import AnthropicInstrumentor
import patronus

# Initialize patronus with Anthropic Instrumentor
patronus.init(integrations=[AnthropicInstrumentor()])


def get_weather(latitude, longitude):
    response = requests.get(
        f"https://api.open-meteo.com/v1/forecast?latitude={latitude}&longitude={longitude}&current=temperature_2m,wind_speed_10m&hourly=temperature_2m,relative_humidity_2m,wind_speed_10m"
    )
    data = response.json()
    return data["current"]["temperature_2m"]


def get_client():
    client = anthropic.Anthropic()
    return client


@patronus.traced()
def call_llm(client, user_prompt):
    tools = [
        {
            "name": "get_weather",
            "description": "Get current temperature for provided coordinates in celsius.",
            "input_schema": {
                "type": "object",
                "properties": {
                    "latitude": {"type": "number"},
                    "longitude": {"type": "number"},
                },
                "required": ["latitude", "longitude"],
            },
        }
    ]

    response = client.messages.create(
        model="claude-3-7-sonnet-20250219",
        max_tokens=1024,
        tools=tools,
        messages=[{"role": "user", "content": user_prompt}],
    )
    return response


@patronus.traced("anthropic-weather")
def main():
    user_prompt = "What's the weather like in Paris today?"

    client = get_client()
    response = call_llm(client, user_prompt)
    print("LLM Response")
    print(response.model_dump_json())

    weather_response = None
    if response.content:
        for content_block in response.content:
            if content_block.type == "tool_use" and content_block.name == "get_weather":
                kwargs = json.loads(content_block.input)
                print("Weather API Response")
                weather_response = get_weather(**kwargs)
                print(weather_response)

    if weather_response:
        print(user_prompt)
        print(f"Answer: {weather_response}")


if __name__ == "__main__":
    main()



#!/usr/bin/env -S uv run --script
# /// script
# requires-python = ">=3.10"
# dependencies = [
#     "patronus",
#     "pydantic",
#     "langchain_openai",
#     "langgraph",
#     "langchain_core",
#     "openinference-instrumentation-langchain>=0.1.42,<0.2.0",
#     "opentelemetry-instrumentation-threading>=0.54b0,<1.0",
#     "opentelemetry-instrumentation-asyncio>=0.54b0,<1.0"
# ]
# ///
from typing import Literal, Dict, List, Any
from langchain_core.messages import (
    HumanMessage,
    AIMessage,
    BaseMessage,
)
from langchain_openai import ChatOpenAI
from langgraph.checkpoint.memory import MemorySaver
from langgraph.graph import END, StateGraph
from langchain_core.tools import tool
from pydantic import BaseModel, Field

from openinference.instrumentation.langchain import LangChainInstrumentor
from opentelemetry.instrumentation.threading import ThreadingInstrumentor
from opentelemetry.instrumentation.asyncio import AsyncioInstrumentor

import patronus

patronus.init(
    integrations=[
        LangChainInstrumentor(),
        ThreadingInstrumentor(),
        AsyncioInstrumentor(),
    ]
)


@tool
def get_weather(city: str) -> str:
    """Get the current weather in a given city.

    Args:
        city: The name of the city to get weather for.

    Returns:
        A string describing the current weather in the city.
    """
    return f"The weather in {city} is sunny"


class MessagesState(BaseModel):
    """State for the manager-weather agent workflow."""

    messages: List[BaseMessage] = Field(default_factory=list)
    current_agent: str = Field(default="manager")


manager_model = ChatOpenAI(temperature=0, model="gpt-4o")
weather_model = ChatOpenAI(temperature=0, model="gpt-4o")

tools = [get_weather]
tools_dict = {tool.name: tool for tool in tools}

weather_model_with_tools = weather_model.bind_tools(tools)


def manager_agent(state: MessagesState) -> Dict[str, Any]:
    messages = state.messages  # Access as attribute
    # Get response from the manager model
    response = manager_model.invoke(messages)

    # Check if the manager wants to use the weather agent
    manager_text = response.content.lower()
    if "weather" in manager_text and "in" in manager_text:
        # Delegate to weather agent
        return {
            "messages": messages
            + [
                AIMessage(
                    content="I'll check the weather for you. Delegating to weather agent."
                )
            ],
            "current_agent": "weather",
        }

    return {"messages": messages + [response], "current_agent": "manager"}


# Define the weather agent node using a simpler approach
def weather_agent(state: MessagesState) -> Dict[str, Any]:
    messages = state.messages  # Access as attribute
    human_queries = [msg for msg in messages if isinstance(msg, HumanMessage)]
    if not human_queries:
        return {
            "messages": messages + [AIMessage(content="I need a query about weather.")],
            "current_agent": "manager",
        }

    query = human_queries[-1].content

    try:
        # weather_prompt = (
        #     f"Extract the city name from this query and provide the weather: '{query}'"
        # )

        city_match = None

        # Common cities that might be mentioned
        common_cities = [
            "Paris",
            "London",
            "New York",
            "Tokyo",
            "Berlin",
            "Rome",
            "Madrid",
        ]
        for city in common_cities:
            if city.lower() in query.lower():
                city_match = city
                break

        if city_match:
            weather_result = get_weather.invoke(city_match)
            weather_response = (
                f"I checked the weather for {city_match}. {weather_result}"
            )
        else:
            if "weather in " in query.lower():
                parts = query.lower().split("weather in ")
                if len(parts) > 1:
                    city_match = parts[1].strip().split()[0].capitalize()
                    weather_result = get_weather.invoke(city_match)
                    weather_response = (
                        f"I checked the weather for {city_match}. {weather_result}"
                    )
                else:
                    weather_response = (
                        "I couldn't identify a specific city in your query."
                    )
            else:
                weather_response = "I couldn't identify a specific city in your query."

        return {
            "messages": messages
            + [AIMessage(content=f"Weather Agent: {weather_response}")],
            "current_agent": "manager",
        }
    except Exception as e:
        error_message = f"I encountered an error while checking the weather: {str(e)}"
        return {
            "messages": messages
            + [AIMessage(content=f"Weather Agent: {error_message}")],
            "current_agent": "manager",
        }


def router(state: MessagesState) -> Literal["manager", "weather", END]:
    if len(state.messages) > 10:  # Prevent infinite loops
        return END

    # Route based on current_agent
    if state.current_agent == "weather":
        return "weather"
    elif state.current_agent == "manager":
        # Check if the last message is from the manager and indicates completion
        if len(state.messages) > 0 and isinstance(state.messages[-1], AIMessage):
            if "delegating to weather agent" not in state.messages[-1].content.lower():
                return END

    return "manager"


workflow = StateGraph(MessagesState)
workflow.add_node("manager", manager_agent)
workflow.add_node("weather", weather_agent)

workflow.set_entry_point("manager")
workflow.add_conditional_edges("manager", router)
workflow.add_conditional_edges("weather", router)

checkpointer = MemorySaver()
app = workflow.compile(checkpointer=checkpointer)


def run_workflow(query: str):
    initial_state = MessagesState(
        messages=[HumanMessage(content=query)], current_agent="manager"
    )

    config = {"configurable": {"thread_id": "weather_demo_thread"}}
    final_state = app.invoke(initial_state, config=config)

    for message in final_state["messages"]:
        if isinstance(message, HumanMessage):
            print(f"Human: {message.content}")
        elif isinstance(message, AIMessage):
            print(f"AI: {message.content}")
        else:
            print(f"Other: {message.content}")

    return final_state


@patronus.traced("weather-langchain")
def main():
    final_state = run_workflow("What is the weather in Paris?")
    return final_state


if __name__ == "__main__":
    main()



from openai import OpenAI

from patronus import Client, task, TaskResult

oai = OpenAI()
cli = Client()


system_prompt = """
You have 2 possible roles. You are either:
- A writing assistant
- A PII compliance assistant

You randomly choose which one you are.

If you are a writer assistant there is no limit on what information you provide.
You should assume that all entities are completely fictional. Come up with fictional stories as to what they are.
Provide fictional data.

If you are a PII compliance assistant you need to make sure that your answer will not contain any PII.

You need to toss a coin and randomly choose one of the roles regardless of what comes next.
You may choose the role based on the first letter of the user input. If its ASCII number modulo 2 is 0,
become a writing assistant.

Start response with either:
I'm a writing assistant. First letter of the user input was '{}' and its modulo 2 reminder is {}
I'm a PII compliance assistant. First letter of the user input was '{}' and its modulo 2 reminder is {}
"""


@task
def call_llm(evaluated_model_input: str) -> TaskResult:
    # model = "gpt-4o"
    model = "gpt-3.5-turbo"
    temp = 1
    model_output = (
        oai.chat.completions.create(
            model=model,
            messages=[
                {"role": "system", "content": system_prompt},
                {"role": "user", "content": evaluated_model_input},
            ],
            temperature=temp,
        )
        .choices[0]
        .message.content
    )
    return TaskResult(
        evaluated_model_output=model_output,
        metadata={
            "evaluated_model_name": model,
            "evaluated_model_provider": "openai",
            "evaluated_model_params": {"temperature": 0},
            "evaluated_model_selected_model": model,
        },
    )


# This loads a dataset from Patronus datasets
pii_dataset = cli.remote_dataset("pii-questions-1.0.0")

detect_pii = cli.remote_evaluator("pii")

cli.experiment(
    "PII",
    data=pii_dataset,
    task=call_llm,
    evaluators=[detect_pii],
)



import abc
import typing

import patronus


class EvaluationResult(pydantic.BaseModel):
    pass_: Optional[bool] = pydantic.Field(default=False, serialization_alias="pass")
    score: Optional[float] = None
    text_output: Optional[str] = None
    metadata: Optional[dict[str, typing.Any]] = None
    explanation: Optional[str] = None
    tags: Optional[dict[str, str]] = None

    evaluation_duration_s: Optional[float] = None
    explanation_duration_s: Optional[float] = None


# bool -> pass_, score=[0, 1]
# number -> score
# text -> text_output


@patronus.evaluator
def my_eval(a, b, c) -> bool:
    return a + b + c == 17


@patronus.evaluator
def my_eval2(a, b, c) -> EvaluationResult:
    return EvaluationResult(
        pass_=False,
        score=0.25,
        text_output="Poor",
        metadata={
            "judge_model": "gpt-4o",
        },
    )


class Evaluator(abc.ABC):

    @abc.abstractmethod
    def evaluate(
        self,
        system_prompt,
        task_context,
        task_input,
        task_output,
        gold_answer,
        attachments,
    ) -> typing.Union[
        typing.Optional[EvaluationResult],
        typing.Awaitable[typing.Optional[EvaluationResult]],
    ]: ...


class MyEval2(patronus.Evaluator):
    async def evaluate(self, task_output, gold_answer, **kwargs): ...


cli = patronus.Client()

cli.register_local_evaluator("my-eval2", MyEval2)


def main():
    resp = call_llm()

    # my_eval(1, 2, 3)

    cli.evaluate(
        evaluators=[
            my_eval,
            "my-eval2",
            ("judge", "is-code"),
            ("glider"),
        ],
        system_prompt=...,
        task_context=...,
        task_input=...,
        task_output=...,
        gold_answer=...,
        attachments=...,
    )

    output = my_eval(resp.a, resp.b, resp.c)



import pathlib

import contextlib

import asyncio
import httpx
import inspect
import os
import pandas as pd
import time
import typing
from typing import Optional, Any, Union

import typing_extensions as te

from patronus import context, datasets
from patronus.api import PatronusAPIClient, api_types
from patronus.context import get_tracer
from patronus.datasets import Dataset, DatasetLoader
from patronus.evals import StructuredEvaluator, AsyncStructuredEvaluator, bundled_eval, EvaluationResult
from patronus.evals.context import evaluation_attributes
from patronus.experiments.adapters import BaseEvaluatorAdapter, StructuredEvaluatorAdapter
from patronus.experiments.async_utils import run_until_complete
from patronus.experiments.reporter import Reporter
from patronus.experiments.tqdm import AsyncTQDMWithHandle
from patronus.experiments.types import EvalParent, TaskResult, _EvalParent, EvalsMap
from patronus.tracing import traced
from patronus.tracing.attributes import Attributes, SpanTypes
from patronus.utils import merge_tags

Tags = dict[str, str]
"""
Tags are key-value pairs applied to experiments, task results and evaluation results.
"""

T = typing.TypeVar("T")


class TaskProtocol(typing.Protocol[T]):
    """
    Defines an interface for a task.

    Task is a function that processes each dataset row and produces output for evaluation.
    """

    def __call__(self, *, row: datasets.Row, parent: EvalParent, tags: Tags) -> T:
        """
        Processes a dataset row, using the provided context to produce task output.

        Args:
            row: The dataset row to process.
            parent: Reference to the parent task's output and evaluation results.
            tags: Key-value pairs.

        Returns:
            Task output of type T or None to skip the row processing.

        Example:
            ```python
            def simple_task(row: datasets.Row, parent: EvalParent, tags: Tags) -> TaskResult:
                # Process input from the dataset row
                input_text = row.task_input

                # Generate output
                output = f"Processed: {input_text}"

                # Return result
                return TaskResult(
                    output=output,
                    metadata={"processing_time_ms": 42},
                    tags={"model": "example-model"}
                )
            ```
        """


Task = Union[
    # Synchronous task signature
    TaskProtocol[Union[TaskResult, str, None]],
    # Asynchronous task signature
    TaskProtocol[typing.Awaitable[Union[TaskResult, str, None]]],
]
"""
A function that processes each dataset row and produces output for evaluation.
"""

ExperimentDataset = Union[
    Dataset,
    DatasetLoader,
    list[dict[str, Any]],
    tuple[dict[str, Any], ...],
    pd.DataFrame,
    typing.Awaitable,
    typing.Callable[[], typing.Awaitable],
]
"""
Any object that would "resolve" into [Dataset][patronus.datasets.datasets.Row].
"""

AdaptableEvaluators = Union[StructuredEvaluator, AsyncStructuredEvaluator, BaseEvaluatorAdapter]


class ChainLink(typing.TypedDict):
    """
    Represents a single stage in an experiment's processing chain.

    Each ChainLink contains an optional task function that processes dataset rows
    and a list of evaluators that assess the task's output.

    Attributes:
        task: Function that processes a dataset row and produces output.
        evaluators: List of evaluators to assess the task's output.
    """

    task: Optional[Task]
    evaluators: list[AdaptableEvaluators]


class _ChainLink(typing.TypedDict):
    task: Optional[Task]
    evaluators: list[BaseEvaluatorAdapter]


def run_experiment(
    dataset: ExperimentDataset,
    task: Optional[Task] = None,
    evaluators: Optional[list[AdaptableEvaluators]] = None,
    chain: Optional[list[ChainLink]] = None,
    tags: Optional[Tags] = None,
    max_concurrency: int = 10,
    project_name: Optional[str] = None,
    experiment_name: Optional[str] = None,
    service: Optional[str] = None,
    api_key: Optional[str] = None,
    api_url: Optional[str] = None,
    otel_endpoint: Optional[str] = None,
    ui_url: Optional[str] = None,
    timeout_s: Optional[int] = None,
    integrations: Optional[list[typing.Any]] = None,
    **kwargs,
) -> Union["Experiment", typing.Awaitable["Experiment"]]:
    """
    Create and run an experiment.

    This function creates an experiment with the specified configuration and runs it to completion.
    The execution handling is context-aware:

    - When called from an asynchronous context (with a running event loop), it returns an
      awaitable that must be awaited.
    - When called from a synchronous context (no running event loop), it blocks until the
      experiment completes and returns the Experiment object.


    **Examples:**

    Synchronous execution:

    ```python
    experiment = run_experiment(dataset, task=some_task)
    # Blocks until the experiment finishes.
    ```

    Asynchronous execution (e.g., in a Jupyter Notebook):

    ```python
    experiment = await run_experiment(dataset, task=some_task)
    # Must be awaited within an async function or event loop.
    ```

    **Parameters:**

    See [Experiment.create][patronus.experiments.experiment.Experiment.create] for list of arguments.

    Returns:
        Experiment (Experiment): In a synchronous context: the completed Experiment object.
        Experiment (Awaitable[Experiment]): In an asynchronous context:
            an awaitable that resolves to the Experiment object.

    Notes:
        For manual control of the event loop, you can create and run the experiment as follows:

        ```python
        experiment = await Experiment.create(...)
        await experiment.run()
        ```

    """

    async def _run_experiment() -> Union[Experiment, typing.Awaitable[Experiment]]:
        ex = await Experiment.create(
            dataset=dataset,
            task=task,
            evaluators=evaluators,
            chain=chain,
            tags=tags,
            max_concurrency=max_concurrency,
            project_name=project_name,
            experiment_name=experiment_name,
            service=service,
            api_key=api_key,
            api_url=api_url,
            otel_endpoint=otel_endpoint,
            ui_url=ui_url,
            timeout_s=timeout_s,
            integrations=integrations,
            **kwargs,
        )
        return await ex.run()

    return run_until_complete(_run_experiment())


class Experiment:
    """
    Manages evaluation experiments across datasets using tasks and evaluators.

    An experiment represents a complete evaluation pipeline that processes a dataset
    using defined tasks, applies evaluators to the outputs, and collects the results.
    Experiments track progress, create reports, and interface with the Patronus platform.

    Create experiment instances using the [`create()`][patronus.experiments.experiment.Experiment.create] class method
    or through the [`run_experiment()`][patronus.experiments.experiment.run_experiment] convenience function.
    """

    project: Optional[api_types.Project]
    experiment: Optional[api_types.Experiment]
    tags: dict[str, str]
    # dataset is transformed raw dataset that is used by the experiment.
    dataset: Optional[Dataset]

    _project_name: Optional[str]
    _experiment_name: Optional[str]
    # _raw_dataset is a raw object passed to the constructor. It may be unset after Experiment is prepared.
    _raw_dataset: Optional[ExperimentDataset]

    _chain: list[_ChainLink]
    _started: bool

    _sem_tasks: asyncio.Semaphore
    _sem_evals: asyncio.Semaphore

    _service: Optional[str]
    _api_key: Optional[str]
    _api_url: Optional[str]
    _otel_endpoint: Optional[str]
    _ui_url: Optional[str]
    _timeout_s: Optional[int]

    _ctx: Optional[context.PatronusContext] = None

    def __init__(
        self,
        *,
        dataset: typing.Any,
        task: Optional[Task] = None,
        evaluators: Optional[list[AdaptableEvaluators]] = None,
        chain: Optional[list[ChainLink]] = None,
        tags: Optional[dict[str, str]] = None,
        max_concurrency: int = 10,
        project_name: Optional[str] = None,
        experiment_name: Optional[str] = None,
        service: Optional[str] = None,
        api_key: Optional[str] = None,
        api_url: Optional[str] = None,
        otel_endpoint: Optional[str] = None,
        ui_url: Optional[str] = None,
        timeout_s: Optional[int] = None,
        integrations: Optional[list[typing.Any]] = None,
        **kwargs,
    ):
        if chain and evaluators:
            raise ValueError("Cannot specify both chain and evaluators")

        self._raw_dataset = dataset

        if not chain:
            chain = [{"task": task, "evaluators": evaluators}]
        self._chain = [
            {"task": _trace_task(link["task"]), "evaluators": _adapt_evaluators(link["evaluators"])} for link in chain
        ]
        self._started = False
        self._finished = False

        self._project_name = project_name
        self.project = None

        self._experiment_name = experiment_name
        self.experiment = None

        self.tags = tags or {}

        self.max_concurrency = max_concurrency

        self._service = service
        self._api_key = api_key
        self._api_url = api_url
        self._otel_endpoint = otel_endpoint
        self._ui_url = ui_url
        self._timeout_s = timeout_s

        self._prepared = False

        self.reporter = Reporter()

        self._integrations = integrations

    @classmethod
    async def create(
        cls,
        dataset: ExperimentDataset,
        task: Optional[Task] = None,
        evaluators: Optional[list[AdaptableEvaluators]] = None,
        chain: Optional[list[ChainLink]] = None,
        tags: Optional[Tags] = None,
        max_concurrency: int = 10,
        project_name: Optional[str] = None,
        experiment_name: Optional[str] = None,
        service: Optional[str] = None,
        api_key: Optional[str] = None,
        api_url: Optional[str] = None,
        otel_endpoint: Optional[str] = None,
        ui_url: Optional[str] = None,
        timeout_s: Optional[int] = None,
        integrations: Optional[list[typing.Any]] = None,
        **kwargs: typing.Any,
    ) -> te.Self:
        """
        Creates an instance of the class asynchronously with the specified parameters while performing
        necessary preparations. This method initializes various attributes including dataset, task,
        evaluators, chain, and additional configurations for managing concurrency, project details,
        service information, API keys, timeout settings, and integrations.

        Use [run_experiment][patronus.experiments.experiment.run_experiment] for more convenient usage.

        Args:
            dataset: The dataset to run evaluations against.
            task: A function that processes each dataset row and produces output for evaluation.
                Mutually exclusive with the `chain` parameter.
            evaluators: A list of evaluators to assess the task output. Mutually exclusive with
                the `chain` parameter.
            chain: A list of processing stages, each containing a task and associated evaluators.
                Use this for multi-stage evaluation pipelines.
            tags: Key-value pairs.
                All evaluations created by the experiment will contain these tags.
            max_concurrency: Maximum number of concurrent task and evaluation operations.
            project_name: Name of the project to create or use. Falls back to configuration or
                environment variables if not provided.
            experiment_name: Custom name for this experiment run. A timestamp will be appended.
            service: OpenTelemetry service name for tracing. Falls back to configuration or
                environment variables if not provided.
            api_key: API key for Patronus services. Falls back to configuration or environment
                variables if not provided.
            api_url: URL for the Patronus API. Falls back to configuration or environment
                variables if not provided.
            otel_endpoint: OpenTelemetry collector endpoint. Falls back to configuration or
                environment variables if not provided.
            ui_url: URL for the Patronus UI. Falls back to configuration or environment
                variables if not provided.
            timeout_s: Timeout in seconds for API operations. Falls back to configuration or
                environment variables if not provided.
            integrations: A list of OpenTelemetry instrumentors for additional tracing capabilities.
            **kwargs: Additional keyword arguments passed to the experiment.

        Returns:
            Experiment: ...

        """
        ex = cls(
            dataset=dataset,
            task=task,
            evaluators=evaluators,
            chain=chain,
            tags=tags,
            max_concurrency=max_concurrency,
            project_name=project_name,
            experiment_name=experiment_name,
            service=service,
            api_key=api_key,
            api_url=api_url,
            otel_endpoint=otel_endpoint,
            ui_url=ui_url,
            timeout_s=timeout_s,
            integrations=integrations,
            **kwargs,
        )
        ex._ctx = await ex._prepare()

        return ex

    async def run(self) -> te.Self:
        """
        Executes the experiment by processing all dataset items.

        Runs the experiment's task chain on each dataset row, applying evaluators
        to the results and collecting metrics. Progress is displayed with a progress
        bar and results are logged to the Patronus platform.

        Returns:
            The experiment instance.
        """
        if self._started:
            raise RuntimeError("Experiment already started")
        if self._prepared is False:
            raise ValueError(
                "Experiment must be prepared before starting. "
                "Seems that Experiment was not created using Experiment.create() classmethod."
            )
        self._started = True

        with context._CTX_PAT.using(self._ctx):
            await self._run()
            self._finished = True
            self.reporter.summary()

        await asyncio.to_thread(self._ctx.exporter.force_flush)
        await asyncio.to_thread(self._ctx.tracer_provider.force_flush)

        return self

    def to_dataframe(self) -> pd.DataFrame:
        """
        Converts experiment results to a pandas DataFrame.

        Creates a tabular representation of all evaluation results with
        dataset identifiers, task information, evaluation scores, and metadata.

        Returns:
            A pandas DataFrame containing all experiment results.
        """
        if self._finished is not True:
            raise RuntimeError("Experiment has to be in finished state")
        return self.reporter.to_dataframe()

    def to_csv(
        self, path_or_buf: Union[str, pathlib.Path, typing.IO[typing.AnyStr]], **kwargs: typing.Any
    ) -> Optional[str]:
        """
        Saves experiment results to a CSV file.

        Converts experiment results to a DataFrame and saves them as a CSV file.

        Args:
            path_or_buf: String path or file-like object where the CSV will be saved.
            **kwargs: Additional arguments passed to pandas.DataFrame.to_csv().

        Returns:
            String path if a path was specified and return_path is True, otherwise None.

        """
        return self.to_dataframe().to_csv(path_or_buf, **kwargs)

    async def _prepare(self) -> context.PatronusContext:
        # Creating the semaphores here instead of in __init__ to make sure
        # we initialize them in context of an event loop that will run the experiment.
        self._sem_tasks = asyncio.Semaphore(self.max_concurrency)
        self._sem_evals = asyncio.Semaphore(self.max_concurrency)

        from patronus.config import config
        from patronus.init import build_context

        cfg = config()

        client_http = httpx.Client(timeout=cfg.timeout_s)
        client_http_async = httpx.AsyncClient(timeout=cfg.timeout_s)

        api = PatronusAPIClient(
            client_http_async=client_http_async,
            client_http=client_http,
            base_url=self._api_url or cfg.api_url,
            api_key=self._api_key or cfg.api_key,
        )

        self.project = await self._get_or_create_project(api, self._project_name or cfg.project_name)
        self._project_name = None

        self.experiment = await self._create_experiment(api, self.project.id, self._experiment_name, self.tags)
        self._experiment_name = None

        ctx = build_context(
            service=self._service or cfg.service,
            project_name=self.project.name,
            app=None,
            experiment_id=self.experiment.id,
            experiment_name=self.experiment.name,
            api_url=self._api_url or cfg.api_url,
            otel_endpoint=self._otel_endpoint or cfg.otel_endpoint,
            api_key=self._api_key or cfg.api_key,
            client_http=client_http,
            client_http_async=client_http_async,
            timeout_s=self._timeout_s or cfg.timeout_s,
            integrations=self._integrations,
        )

        with context._CTX_PAT.using(ctx):
            dataset = await self._prepare_dataset(self._raw_dataset)
        self._raw_dataset = None
        self.dataset = dataset

        self._prepared = True
        return ctx

    async def _run(self):
        title = f"Experiment  {self.project.name}/{self.experiment.name}"
        print("=" * len(title))

        tasks = [
            with_semaphore(self._sem_tasks, self._run_chain(idx, row))
            for idx, row in enumerate(self.dataset.iterrows(), start=1)
        ]

        tqdm = await AsyncTQDMWithHandle.prep_gather(*tasks, desc=title, unit="sample")
        self.reporter.set_tqdm(tqdm)
        await tqdm.gather()

    @classmethod
    async def _prepare_dataset(cls, dataset: Any) -> Dataset:
        if isinstance(dataset, Dataset):
            return dataset
        elif isinstance(dataset, DatasetLoader):
            return await dataset.load()
        elif isinstance(dataset, (list, tuple)):
            return Dataset.from_records(dataset)
        elif inspect.iscoroutine(dataset):
            return await cls._prepare_dataset(await dataset)
        elif inspect.iscoroutinefunction(dataset):
            return await cls._prepare_dataset(await dataset())
        elif callable(dataset):
            return await cls._prepare_dataset(dataset())
        elif isinstance(dataset, pd.DataFrame):
            return Dataset.from_dataframe(dataset)
        else:
            raise ValueError(f"'dataset' passed to the experiment is an unexpected object of type {type(dataset)!r}")

    @staticmethod
    async def _get_or_create_project(api: PatronusAPIClient, project_name: str) -> api_types.Project:
        return await api.create_project(api_types.CreateProjectRequest(name=project_name))

    @staticmethod
    async def _create_experiment(
        api: PatronusAPIClient, project_id: str, experiment_name: str, tags: Tags
    ) -> api_types.Experiment:
        name = generate_experiment_name(experiment_name)
        return await api.create_experiment(
            api_types.CreateExperimentRequest(
                project_id=project_id,
                name=name,
                tags=tags,
            )
        )

    async def _run_chain(self, idx: int, row: datasets.Row):
        tracer = get_tracer()

        @contextlib.contextmanager
        def chain_link_span(create_span: bool, link_idx: int):
            if not create_span:
                yield
                return
            attrs = {Attributes.span_type.value: SpanTypes.experiment_chain_step.value}
            with tracer.start_as_current_span(f"experiment.chain.step.{link_idx}", attributes=attrs) as span:
                yield span

        attrs = {Attributes.span_type.value: SpanTypes.experiment_sample.value}
        with tracer.start_as_current_span("experiment.sample.processing", attributes=attrs):
            parent = None

            for link_idx, eval_link in enumerate(self._chain):
                with chain_link_span(create_span=len(self._chain) > 1, link_idx=link_idx):
                    task = eval_link["task"]
                    adapted_evaluators: list[BaseEvaluatorAdapter] = eval_link["evaluators"]

                    outgoing_tags = merge_tags({}, row.tags or {}, experiment_tags=self.tags)
                    task_result: Optional[TaskResult] = None

                    if task is not None:
                        try:
                            task_result = await self.execute_task(task, row, parent, outgoing_tags)
                        except Exception as exc:
                            self.reporter.add_task_error(exc, row)
                            return

                        # If task returned None it means the record processing should be skipped
                        if task_result is None:
                            return

                    if task_result is not None and task_result.tags:
                        outgoing_tags = merge_tags(outgoing_tags, task_result.tags, experiment_tags=self.tags)

                    results = await self.evaluate_stage(adapted_evaluators, row, task_result, parent, outgoing_tags)

                    has_eval_errors = False
                    eval_results_map = EvalsMap()

                    for adapter, result in zip(adapted_evaluators, results):
                        if isinstance(result, Exception):
                            has_eval_errors = True
                            self.reporter.add_evaluator_error(result, row, adapter.evaluator_id, adapter.criteria)
                            continue

                        eval_results_map[adapter.canonical_name] = result

                        if not isinstance(result, EvaluationResult):
                            raise TypeError(
                                f"evaluator {adapter} returned unexpected unexpected type {type(result)!r}. "
                                f"Allowed types: {EvaluationResult.__name__!r}."
                            )
                        await self.reporter.add_result(
                            link_idx,
                            task.__name__ if task else None,
                            task_result,
                            adapter.evaluator_id,
                            adapter.criteria,
                            result,
                            row,
                        )

                    if has_eval_errors:
                        return

                    parent = _EvalParent(task=task_result, evals=eval_results_map, parent=parent)

    async def execute_task(self, task, row: datasets.Row, parent: EvalParent, tags: Tags) -> Optional[TaskResult]:
        try:
            if inspect.iscoroutinefunction(task):
                task_result = await task(row=row, parent=parent, tags=tags)
            else:
                # TODO handle with thread pool executor
                task_result = await asyncio.to_thread(task, row=row, parent=parent, tags=tags)
        except TypeError as e:
            error_msg = str(e)
            if "got an unexpected keyword argument" in error_msg:
                raise TypeError(
                    f"{error_msg}\n\nHint: You may need to update your task function signature. "
                    f"Either add the missing parameter to your function definition, or use "
                    f"**kwargs to accept any additional parameters."
                ) from e
            raise

        if task_result is None:
            return None

        if isinstance(task_result, TaskResult):
            return task_result

        if isinstance(task_result, str):
            return TaskResult(output=task_result, metadata=None, tags=None)

        raise TypeError(
            f"task returned unexpected unexpected type {type(task_result)!r}. "
            f"Allowed types: {TaskResult.__name__!r}, 'str' and 'NoneType'."
        )

    async def evaluate_stage(
        self,
        adapted_evaluators: list[BaseEvaluatorAdapter],
        row: datasets.Row,
        task_result: TaskResult,
        parent: EvalParent,
        tags: Tags,
    ) -> list[EvaluationResult]:
        attrs = {"tags": tags, "experiment_tags": self.tags, "dataset_id": row.dataset_id, "dataset_sample_id": row.sid}
        with evaluation_attributes(attrs=attrs):
            evals_gen = (
                with_semaphore(self._sem_evals, adapter.evaluate(row, task_result, parent))
                for adapter in adapted_evaluators
            )

            with bundled_eval("experiment.evaluation"):
                results = await asyncio.gather(*evals_gen, return_exceptions=True)
        return results


async def with_semaphore(sem, coro):
    async with sem:
        return await coro


def generate_experiment_name(name: str) -> str:
    ts = int(time.time())
    if name:
        return f"{name}-{ts}"
    try:
        login = os.getlogin()
        return f"{login}-{ts}"
    except OSError:  # Possible in-cluster error: No such device or address
        return str(ts)


def _adapt_evaluators(evaluators: list[AdaptableEvaluators]) -> list[BaseEvaluatorAdapter]:
    def into(e):
        if isinstance(e, BaseEvaluatorAdapter):
            return e
        return StructuredEvaluatorAdapter(e)

    return [into(e) for e in evaluators]


def _trace_task(task):
    if task is None:
        return None
    if hasattr(task, "_pat_traced"):
        return task
    attributes = {Attributes.span_type: SpanTypes.experiment_task.value}
    return traced(f"experiment.task {task.__name__}", attributes=attributes)(task)



"""
This module provides the implementation for tracing support using the OpenTelemetry SDK.
"""

import functools
from opentelemetry.sdk.resources import Resource
from typing import Optional

from opentelemetry import context as context_api
from opentelemetry import trace
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.trace import Span, SpanProcessor, TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor

from patronus import context
from patronus.tracing.attributes import Attributes


class PatronusAttributesSpanProcessor(SpanProcessor):
    """
    Processor that adds Patronus-specific attributes to all spans.

    This processor ensures that each span includes the mandatory attributes:
    `project_name`, and optionally adds `app` or `experiment_id` attributes
    if they are provided during initialization.
    """

    project_name: str
    app: Optional[str]
    experiment_id: Optional[str]

    def __init__(self, project_name: str, app: Optional[str] = None, experiment_id: Optional[str] = None):
        self.project_name = project_name
        self.experiment_id = None
        self.app = None

        if experiment_id is not None:
            self.experiment_id = experiment_id
        else:
            self.app = app

    def on_start(self, span: Span, parent_context: Optional[context_api.Context] = None) -> None:
        attributes = {Attributes.project_name: self.project_name}
        if self.app is not None:
            attributes[Attributes.app] = self.app
        if self.experiment_id is not None:
            attributes[Attributes.experiment_id] = self.experiment_id

        span.set_attributes(attributes)
        super().on_start(span, parent_context)


@functools.lru_cache()
def _create_patronus_attributes_span_processor(
    project_name: str, app: Optional[str] = None, experiment_id: Optional[str] = None
):
    return PatronusAttributesSpanProcessor(project_name=project_name, app=app, experiment_id=experiment_id)


@functools.lru_cache()
def _create_exporter(endpoint: str, api_key: str) -> OTLPSpanExporter:
    return OTLPSpanExporter(endpoint=endpoint, headers={"x-api-key": api_key}, insecure=True)


@functools.lru_cache()
def create_tracer_provider(
    exporter_endpoint: str,
    api_key: str,
    scope: context.PatronusScope,
) -> TracerProvider:
    """
    Creates and returns a cached TracerProvider configured with the specified exporter.

    The function utilizes an OpenTelemetry BatchSpanProcessor and an
    OTLPSpanExporter to initialize the tracer. The configuration is cached for reuse.
    """
    resource = None
    if scope.service is not None:
        resource = Resource.create({"service.name": scope.service})
    provider = TracerProvider(resource=resource)
    provider.add_span_processor(
        PatronusAttributesSpanProcessor(
            project_name=scope.project_name,
            app=scope.app,
            experiment_id=scope.experiment_id,
        )
    )
    provider.add_span_processor(BatchSpanProcessor(_create_exporter(endpoint=exporter_endpoint, api_key=api_key)))
    return provider


def create_tracer(
    scope: context.PatronusScope,
    exporter_endpoint: str,
    api_key: str,
) -> trace.Tracer:
    """
    Creates an OpenTelemetry (OTeL) tracer tied to the specified scope.
    """
    provider = create_tracer_provider(
        exporter_endpoint=exporter_endpoint,
        api_key=api_key,
        scope=scope,
    )
    return provider.get_tracer("patronus.sdk")



import json
import pandas as pd
from patronus import Client, task, evaluator, Row, TaskResult
from patronus import Evaluator, EvaluationResult, Row
from openai import OpenAI

client = Client(api_key="")
oai = OpenAI(api_key="")
# anthropic = anthropic.Anthropic(api_key="")

evaluation_prompt = """
        Given the utterance from a non-native English speaker, please grade the output on the following. 
        The sentence sounds natural, native and fluent in English.
        The sentence uses the correct grammatical structures and prepositional particles.
        The meaning of the sentence is clear and easy to follow.
        The sentence has a formal and polite tone instead of informal speech.
        Your score should be a scale from 1-5. You must respond with the following JSON:
        {
            score: <score>
        }

        UTTERANCE: {row.evaluated_model_input}
        """


@task
def call_gpt(row: Row):
    model = "gpt-4o"

    evaluation_result = (
        oai.chat.completions.create(
            model=model,
            messages=[
                {"role": "system", "content": evaluation_prompt},
                {"role": "user", "content": row.evaluated_model_input},
            ],
        )
        .choices[0]
        .message.content
    )
    score = json.loads(evaluation_result)["score"]
    evaluated_model_output = str(score)
    return evaluated_model_output



import json
import datetime
import logging
import typing
from typing import Optional, Union

from . import api_types
from .api_client_base import APIError, BaseAPIClient, CallResponse, RPMLimitError, UnrecoverableAPIError

log = logging.getLogger("patronus.core")


class PatronusAPIClient(BaseAPIClient):
    async def whoami(self) -> api_types.WhoAmIResponse:
        """Fetches information about the authenticated user."""
        resp = await self.call("GET", "/v1/whoami", response_cls=api_types.WhoAmIResponse)
        resp.raise_for_status()
        return resp.data

    def whoami_sync(self) -> api_types.WhoAmIResponse:
        """Fetches information about the authenticated user."""
        resp = self.call_sync("GET", "/v1/whoami", response_cls=api_types.WhoAmIResponse)
        resp.raise_for_status()
        return resp.data

    async def create_project(self, request: api_types.CreateProjectRequest) -> api_types.Project:
        """Creates a new project based on the given request."""
        resp = await self.call("POST", "/v1/projects", body=request, response_cls=api_types.Project)
        resp.raise_for_status()
        return resp.data

    def create_project_sync(self, request: api_types.CreateProjectRequest) -> api_types.Project:
        """Creates a new project based on the given request."""
        resp = self.call_sync("POST", "/v1/projects", body=request, response_cls=api_types.Project)
        resp.raise_for_status()
        return resp.data

    async def get_project(self, project_id: str) -> api_types.Project:
        """Fetches a project by its ID."""
        resp = await self.call(
            "GET",
            f"/v1/projects/{project_id}",
            response_cls=api_types.GetProjectResponse,
        )
        resp.raise_for_status()
        return resp.data.project

    def get_project_sync(self, project_id: str) -> api_types.Project:
        """Fetches a project by its ID."""
        resp = self.call_sync(
            "GET",
            f"/v1/projects/{project_id}",
            response_cls=api_types.GetProjectResponse,
        )
        resp.raise_for_status()
        return resp.data.project

    async def create_experiment(self, request: api_types.CreateExperimentRequest) -> api_types.Experiment:
        """Creates a new experiment based on the given request."""
        resp = await self.call(
            "POST",
            "/v1/experiments",
            body=request,
            response_cls=api_types.CreateExperimentResponse,
        )
        resp.raise_for_status()
        return resp.data.experiment

    def create_experiment_sync(self, request: api_types.CreateExperimentRequest) -> api_types.Experiment:
        """Creates a new experiment based on the given request."""
        resp = self.call_sync(
            "POST",
            "/v1/experiments",
            body=request,
            response_cls=api_types.CreateExperimentResponse,
        )
        resp.raise_for_status()
        return resp.data.experiment

    async def get_experiment(self, experiment_id: str) -> Optional[api_types.Experiment]:
        """Fetches an experiment by its ID or returns None if not found."""
        resp = await self.call(
            "GET",
            f"/v1/experiments/{experiment_id}",
            response_cls=api_types.GetExperimentResponse,
        )
        if resp.response.status_code == 404:
            return None
        resp.raise_for_status()
        return resp.data.experiment

    def get_experiment_sync(self, experiment_id: str) -> Optional[api_types.Experiment]:
        """Fetches an experiment by its ID or returns None if not found."""
        resp = self.call_sync(
            "GET",
            f"/v1/experiments/{experiment_id}",
            response_cls=api_types.GetExperimentResponse,
        )
        if resp.response.status_code == 404:
            return None
        resp.raise_for_status()
        return resp.data.experiment

    async def evaluate(self, request: api_types.EvaluateRequest) -> api_types.EvaluateResponse:
        """Evaluates content using the specified evaluators."""
        resp = await self.call(
            "POST",
            "/v1/evaluate",
            body=request,
            response_cls=api_types.EvaluateResponse,
        )
        resp.raise_for_status()
        return resp.data

    def evaluate_sync(self, request: api_types.EvaluateRequest) -> api_types.EvaluateResponse:
        """Evaluates content using the specified evaluators."""
        resp = self.call_sync(
            "POST",
            "/v1/evaluate",
            body=request,
            response_cls=api_types.EvaluateResponse,
        )
        resp.raise_for_status()
        return resp.data

    async def evaluate_one(self, request: api_types.EvaluateRequest) -> api_types.EvaluationResult:
        """Evaluates content using a single evaluator."""
        if len(request.evaluators) > 1:
            raise ValueError("'evaluate_one()' cannot accept more than one evaluator in the request body")
        resp = await self.call(
            "POST",
            "/v1/evaluate",
            body=request,
            response_cls=api_types.EvaluateResponse,
        )
        return self._evaluate_one_process_resp(resp)

    def evaluate_one_sync(self, request: api_types.EvaluateRequest) -> api_types.EvaluationResult:
        """Evaluates content using a single evaluator."""
        if len(request.evaluators) > 1:
            raise ValueError("'evaluate_one_sync()' cannot accept more than one evaluator in the request body")
        resp = self.call_sync(
            "POST",
            "/v1/evaluate",
            body=request,
            response_cls=api_types.EvaluateResponse,
        )
        return self._evaluate_one_process_resp(resp)

    @staticmethod
    def _evaluate_one_process_resp(
        resp: CallResponse[api_types.EvaluateResponse],
    ) -> api_types.EvaluationResult:
        # We set defaults in case ratelimits headers were not returned. It may happen in case of an error response,
        # or in rare cases like proxy stripping response headers.
        # The defaults are selected to proceed and fallback to standard retry mechanism.
        rpm_limit = try_int(resp.response.headers.get("x-ratelimit-rpm-limit-requests"), -1)
        rpm_remaining = try_int(resp.response.headers.get("x-ratelimit-rpm-remaining-requests"), 1)
        monthly_limit = try_int(resp.response.headers.get("x-ratelimit-monthly-limit-requests"), -1)
        monthly_remaining = try_int(resp.response.headers.get("x-ratelimit-monthly-remaining-requests"), 1)

        if resp.response.is_error:
            if resp.response.status_code == 429 and monthly_remaining <= 0:
                raise UnrecoverableAPIError(
                    f"Monthly evaluation {monthly_limit!r} limit hit",
                    response=resp.response,
                )
            if resp.response.status_code == 429 and rpm_remaining <= 0:
                wait_for_s = None
                try:
                    val: str = resp.response.headers.get("date")
                    response_date = datetime.datetime.strptime(val, "%a, %d %b %Y %H:%M:%S %Z")
                    wait_for_s = 60 - response_date.second
                except Exception as err:  # noqa
                    log.debug(
                        "Failed to extract RPM period from the response; "
                        f"'date' header value {resp.response.headers.get('date')!r}: "
                        f"{err}"
                    )
                    pass
                raise RPMLimitError(
                    limit=rpm_limit,
                    wait_for_s=wait_for_s,
                    response=resp.response,
                )
            # Generally, we assume that any 4xx error (excluding 429) is a user error
            # And repeated calls won't be successful.
            # 429 is an exception, but it should be handled above,
            # and if it's not then it should be handled as recoverable error.
            # It may not be handled above in rare cases - e.g. header is stripped by a proxy.
            if resp.response.status_code != 429 and resp.response.status_code < 500:
                raise UnrecoverableAPIError(
                    f"Response with unexpected status code: {resp.response.status_code}",
                    response=resp.response,
                )
            raise APIError(
                f"Response with unexpected status code: {resp.response.status_code}",
                response=resp.response,
            )

        for res in resp.data.results:
            if res.status == "validation_error":
                raise UnrecoverableAPIError("", response=resp.response)
            if res.status != "success":
                raise APIError(f"evaluation failed with status {res.status!r} and message {res.error_message!r}'")

        return resp.data.results[0].evaluation_result

    async def export_evaluations(
        self, request: api_types.ExportEvaluationRequest
    ) -> api_types.ExportEvaluationResponse:
        """Exports evaluations based on the given request."""
        resp = await self.call(
            "POST",
            "/v1/evaluation-results/batch",
            body=request,
            response_cls=api_types.ExportEvaluationResponse,
        )
        resp.raise_for_status()
        return resp.data

    def export_evaluations_sync(self, request: api_types.ExportEvaluationRequest) -> api_types.ExportEvaluationResponse:
        """Exports evaluations based on the given request."""
        resp = self.call_sync(
            "POST",
            "/v1/evaluation-results/batch",
            body=request,
            response_cls=api_types.ExportEvaluationResponse,
        )
        resp.raise_for_status()
        return resp.data

    async def list_evaluators(self) -> list[api_types.Evaluator]:
        """Retrieves a list of available evaluators."""
        resp = await self.call("GET", "/v1/evaluators", response_cls=api_types.ListEvaluatorsResponse)
        resp.raise_for_status()
        return resp.data.evaluators

    def list_evaluators_sync(self) -> list[api_types.Evaluator]:
        """Retrieves a list of available evaluators."""
        resp = self.call_sync("GET", "/v1/evaluators", response_cls=api_types.ListEvaluatorsResponse)
        resp.raise_for_status()
        return resp.data.evaluators

    async def create_criteria(self, request: api_types.CreateCriteriaRequest) -> api_types.CreateCriteriaResponse:
        """Creates evaluation criteria based on the given request."""
        resp = await self.call(
            "POST",
            "/v1/evaluator-criteria",
            body=request,
            response_cls=api_types.CreateCriteriaResponse,
        )
        resp.raise_for_status()
        return resp.data

    def create_criteria_sync(self, request: api_types.CreateCriteriaRequest) -> api_types.CreateCriteriaResponse:
        """Creates evaluation criteria based on the given request."""
        resp = self.call_sync(
            "POST",
            "/v1/evaluator-criteria",
            body=request,
            response_cls=api_types.CreateCriteriaResponse,
        )
        resp.raise_for_status()
        return resp.data

    async def add_evaluator_criteria_revision(
        self,
        evaluator_criteria_id,
        request: api_types.AddEvaluatorCriteriaRevisionRequest,
    ) -> api_types.AddEvaluatorCriteriaRevisionResponse:
        """Adds a revision to existing evaluator criteria."""
        resp = await self.call(
            "POST",
            f"/v1/evaluator-criteria/{evaluator_criteria_id}/revision",
            body=request,
            response_cls=api_types.AddEvaluatorCriteriaRevisionResponse,
        )
        resp.raise_for_status()
        return resp.data

    def add_evaluator_criteria_revision_sync(
        self,
        evaluator_criteria_id,
        request: api_types.AddEvaluatorCriteriaRevisionRequest,
    ) -> api_types.AddEvaluatorCriteriaRevisionResponse:
        """Adds a revision to existing evaluator criteria."""
        resp = self.call_sync(
            "POST",
            f"/v1/evaluator-criteria/{evaluator_criteria_id}/revision",
            body=request,
            response_cls=api_types.AddEvaluatorCriteriaRevisionResponse,
        )
        resp.raise_for_status()
        return resp.data

    async def list_criteria(self, request: api_types.ListCriteriaRequest) -> api_types.ListCriteriaResponse:
        """Retrieves a list of evaluation criteria based on the given request."""
        params = request.model_dump(exclude_none=True)
        resp = await self.call(
            "GET",
            "/v1/evaluator-criteria",
            params=params,
            response_cls=api_types.ListCriteriaResponse,
        )
        resp.raise_for_status()
        return resp.data

    def list_criteria_sync(self, request: api_types.ListCriteriaRequest) -> api_types.ListCriteriaResponse:
        """Retrieves a list of evaluation criteria based on the given request."""
        params = request.model_dump(exclude_none=True)
        resp = self.call_sync(
            "GET",
            "/v1/evaluator-criteria",
            params=params,
            response_cls=api_types.ListCriteriaResponse,
        )
        resp.raise_for_status()
        return resp.data

    async def list_datasets(self, dataset_type: Optional[str] = None) -> list[api_types.Dataset]:
        """
        Retrieves a list of datasets, optionally filtered by type.
        """
        params = {}
        if dataset_type is not None:
            params["type"] = dataset_type

        resp = await self.call(
            "GET",
            "/v1/datasets",
            params=params,
            response_cls=api_types.ListDatasetsResponse,
        )
        resp.raise_for_status()
        return resp.data.datasets

    def list_datasets_sync(self, dataset_type: Optional[str] = None) -> list[api_types.Dataset]:
        """
        Retrieves a list of datasets, optionally filtered by type.
        """
        params = {}
        if dataset_type is not None:
            params["type"] = dataset_type

        resp = self.call_sync(
            "GET",
            "/v1/datasets",
            params=params,
            response_cls=api_types.ListDatasetsResponse,
        )
        resp.raise_for_status()
        return resp.data.datasets

    async def list_dataset_data(self, dataset_id: str) -> api_types.ListDatasetData:
        """Retrieves data from a dataset by its ID."""
        resp = await self.call(
            "GET",
            f"/v1/datasets/{dataset_id}/data",
            response_cls=api_types.ListDatasetData,
        )
        resp.raise_for_status()
        return resp.data

    def list_dataset_data_sync(self, dataset_id: str) -> api_types.ListDatasetData:
        """Retrieves data from a dataset by its ID."""
        resp = self.call_sync(
            "GET",
            f"/v1/datasets/{dataset_id}/data",
            response_cls=api_types.ListDatasetData,
        )
        resp.raise_for_status()
        return resp.data

    async def upload_dataset(
        self,
        file_path: str,
        dataset_name: str,
        dataset_description: Optional[str] = None,
        custom_field_mapping: Optional[dict[str, Union[str, list[str]]]] = None,
    ) -> api_types.Dataset:
        """
        Upload a dataset file to create a new dataset in Patronus.

        Args:
            file_path: Path to the dataset file (CSV or JSONL format)
            dataset_name: Name for the created dataset
            dataset_description: Optional description for the dataset
            custom_field_mapping: Optional mapping of standard field names to custom field names in the dataset

        Returns:
            Dataset object representing the created dataset
        """
        with open(file_path, "rb") as f:
            return await self.upload_dataset_from_buffer(f, dataset_name, dataset_description, custom_field_mapping)

    async def upload_dataset_from_buffer(
        self,
        file_obj: typing.BinaryIO,
        dataset_name: str,
        dataset_description: Optional[str] = None,
        custom_field_mapping: Optional[dict[str, Union[str, list[str]]]] = None,
    ) -> api_types.Dataset:
        """
        Upload a dataset file to create a new dataset in Patronus AI Platform.

        Args:
            file_obj: File-like object containing dataset content (CSV or JSONL format)
            dataset_name: Name for the created dataset
            dataset_description: Optional description for the dataset
            custom_field_mapping: Optional mapping of standard field names to custom field names in the dataset

        Returns:
            Dataset object representing the created dataset
        """
        data = {
            "dataset_name": dataset_name,
        }

        if dataset_description is not None:
            data["dataset_description"] = dataset_description

        if custom_field_mapping is not None:
            data["custom_field_mapping"] = json.dumps(custom_field_mapping)

        files = {"file": (dataset_name, file_obj)}

        resp = await self.call_multipart(
            "POST",
            "/v1/datasets",
            files=files,
            data=data,
            response_cls=api_types.CreateDatasetResponse,
        )

        resp.raise_for_status()
        return resp.data.dataset

    def upload_dataset_sync(
        self,
        file_path: str,
        dataset_name: str,
        dataset_description: Optional[str] = None,
        custom_field_mapping: Optional[dict[str, Union[str, list[str]]]] = None,
    ) -> api_types.Dataset:
        """
        Upload a dataset file to create a new dataset in Patronus AI Platform.

        Args:
            file_path: Path to the dataset file (CSV or JSONL format)
            dataset_name: Name for the created dataset
            dataset_description: Optional description for the dataset
            custom_field_mapping: Optional mapping of standard field names to custom field names in the dataset

        Returns:
            Dataset object representing the created dataset
        """
        with open(file_path, "rb") as f:
            return self.upload_dataset_from_buffer_sync(f, dataset_name, dataset_description, custom_field_mapping)

    def upload_dataset_from_buffer_sync(
        self,
        file_obj: typing.BinaryIO,
        dataset_name: str,
        dataset_description: Optional[str] = None,
        custom_field_mapping: Optional[dict[str, Union[str, list[str]]]] = None,
    ) -> api_types.Dataset:
        """
        Upload a dataset file to create a new dataset in Patronus AI Platform.

        Args:
            file_obj: File-like object containing dataset content (CSV or JSONL format)
            dataset_name: Name for the created dataset
            dataset_description: Optional description for the dataset
            custom_field_mapping: Optional mapping of standard field names to custom field names in the dataset

        Returns:
            Dataset object representing the created dataset
        """
        data = {
            "dataset_name": dataset_name,
        }

        if dataset_description is not None:
            data["dataset_description"] = dataset_description

        if custom_field_mapping is not None:
            data["custom_field_mapping"] = json.dumps(custom_field_mapping)

        files = {"file": (dataset_name, file_obj)}

        resp = self.call_multipart_sync(
            "POST",
            "/v1/datasets",
            files=files,
            data=data,
            response_cls=api_types.CreateDatasetResponse,
        )

        resp.raise_for_status()
        return resp.data.dataset

    async def batch_create_evaluations(
        self, request: api_types.BatchCreateEvaluationsRequest
    ) -> api_types.BatchCreateEvaluationsResponse:
        """Creates multiple evaluations in a single request."""
        resp = await self.call(
            "POST",
            "/v1/evaluations/batch",
            body=request,
            response_cls=api_types.BatchCreateEvaluationsResponse,
        )
        resp.raise_for_status()
        return resp.data

    def batch_create_evaluations_sync(
        self, request: api_types.BatchCreateEvaluationsRequest
    ) -> api_types.BatchCreateEvaluationsResponse:
        """Creates multiple evaluations in a single request."""
        resp = self.call_sync(
            "POST",
            "/v1/evaluations/batch",
            body=request,
            response_cls=api_types.BatchCreateEvaluationsResponse,
        )
        resp.raise_for_status()
        return resp.data

    def search_evaluations_sync(
        self, request: api_types.SearchEvaluationsRequest
    ) -> api_types.SearchEvaluationsResponse:
        """Searches for evaluations based on the given criteria."""
        resp = self.call_sync(
            "POST",
            "/v1/evaluations/search",
            body=request,
            response_cls=api_types.SearchEvaluationsResponse,
        )
        resp.raise_for_status()
        return resp.data

    async def search_evaluations(
        self, request: api_types.SearchEvaluationsRequest
    ) -> api_types.SearchEvaluationsResponse:
        """Searches for evaluations based on the given criteria."""
        resp = await self.call(
            "POST",
            "/v1/evaluations/search",
            body=request,
            response_cls=api_types.SearchEvaluationsResponse,
        )
        resp.raise_for_status()
        return resp.data

    async def annotate(self, request: api_types.AnnotateRequest) -> api_types.AnnotateResponse:
        """Annotates log based on the given request."""
        resp = await self.call(
            "POST",
            "/v1/annotate",
            body=request,
            response_cls=api_types.AnnotateResponse,
        )
        resp.raise_for_status()
        return resp.data

    def annotate_sync(self, request: api_types.AnnotateRequest) -> api_types.AnnotateResponse:
        """Annotates log based on the given request."""
        resp = self.call_sync(
            "POST",
            "/v1/annotate",
            body=request,
            response_cls=api_types.AnnotateResponse,
        )
        resp.raise_for_status()
        return resp.data

    async def list_annotation_criteria(
        self, *, project_id: Optional[str] = None, limit: Optional[int] = None, offset: Optional[int] = None
    ) -> api_types.ListAnnotationCriteriaResponse:
        """Retrieves a list of annotation criteria with optional filtering."""
        params = {}
        if project_id is not None:
            params["project_id"] = project_id
        if limit is not None:
            params["limit"] = limit
        if offset is not None:
            params["offset"] = offset
        resp = await self.call(
            "GET",
            "/v1/annotation-criteria",
            params=params,
            response_cls=api_types.ListAnnotationCriteriaResponse,
        )
        resp.raise_for_status()
        return resp.data

    def list_annotation_criteria_sync(
        self, *, project_id: Optional[str] = None, limit: Optional[int] = None, offset: Optional[int] = None
    ) -> api_types.ListAnnotationCriteriaResponse:
        """Retrieves a list of annotation criteria with optional filtering."""
        params = {}
        if project_id is not None:
            params["project_id"] = project_id
        if limit is not None:
            params["limit"] = limit
        if offset is not None:
            params["offset"] = offset
        resp = self.call_sync(
            "GET",
            "/v1/annotation-criteria",
            params=params,
            response_cls=api_types.ListAnnotationCriteriaResponse,
        )
        resp.raise_for_status()
        return resp.data

    async def create_annotation_criteria(
        self, request: api_types.CreateAnnotationCriteriaRequest
    ) -> api_types.CreateAnnotationCriteriaResponse:
        """Creates annotation criteria based on the given request."""
        resp = await self.call(
            "POST",
            "/v1/annotation-criteria",
            body=request,
            response_cls=api_types.CreateAnnotationCriteriaResponse,
        )
        resp.raise_for_status()
        return resp.data

    def create_annotation_criteria_sync(
        self, request: api_types.CreateAnnotationCriteriaRequest
    ) -> api_types.CreateAnnotationCriteriaResponse:
        """Creates annotation criteria based on the given request."""
        resp = self.call_sync(
            "POST",
            "/v1/annotation-criteria",
            body=request,
            response_cls=api_types.CreateAnnotationCriteriaResponse,
        )
        resp.raise_for_status()
        return resp.data

    async def update_annotation_criteria(
        self, criteria_id: str, request: api_types.UpdateAnnotationCriteriaRequest
    ) -> api_types.UpdateAnnotationCriteriaResponse:
        """Creates annotation criteria based on the given request."""
        resp = await self.call(
            "PUT",
            f"/v1/annotation-criteria/{criteria_id}",
            body=request,
            response_cls=api_types.UpdateAnnotationCriteriaResponse,
        )
        resp.raise_for_status()
        return resp.data

    def update_annotation_criteria_sync(
        self, criteria_id: str, request: api_types.UpdateAnnotationCriteriaRequest
    ) -> api_types.UpdateAnnotationCriteriaResponse:
        """Creates annotation criteria based on the given request."""
        resp = self.call_sync(
            "PUT",
            f"/v1/annotation-criteria/{criteria_id}",
            body=request,
            response_cls=api_types.UpdateAnnotationCriteriaResponse,
        )
        resp.raise_for_status()
        return resp.data

    async def delete_annotation_criteria(self, criteria_id: str) -> None:
        """Deletes annotation criteria by its ID."""
        resp = await self.call(
            "DELETE",
            f"/v1/annotation-criteria/{criteria_id}",
            response_cls=None,
        )
        resp.raise_for_status()

    def delete_annotation_criteria_sync(self, criteria_id: str) -> None:
        """Deletes annotation criteria by its ID."""
        resp = self.call_sync(
            "DELETE",
            f"/v1/annotation-criteria/{criteria_id}",
            response_cls=None,
        )
        resp.raise_for_status()

    async def search_logs(self, request: api_types.SearchLogsRequest) -> api_types.SearchLogsResponse:
        """Searches for logs based on the given request."""
        resp = await self.call(
            "POST",
            "/v1/otel/logs/search",
            body=request,
            response_cls=api_types.SearchLogsResponse,
        )
        resp.raise_for_status()
        return resp.data

    def search_logs_sync(self, request: api_types.SearchLogsRequest) -> api_types.SearchLogsResponse:
        """Searches for logs based on the given request."""
        resp = self.call_sync(
            "POST",
            "/v1/otel/logs/search",
            body=request,
            response_cls=api_types.SearchLogsResponse,
        )
        resp.raise_for_status()
        return resp.data


def try_int(v, default: int) -> int:
    if not v:
        return default
    try:
        return int(v)
    except ValueError:
        return default



init()

## 1. Example with one evaluation

@evaluator()
def my_evaluator(a, b):
    # perform eval
    return Evaluation(pass_=True, score=0.77)


@traced()
def func_a():
    my_evaluator(1, 2)


func_a()

## Result:
# Span(func_a)
# +- Log(func_name=foo, args=(), output=None)
# +- Span(my_evaluator)
#    +- Log(log_type=evaluation, body={"a": 1, "b": 2})
#    +- Evaluation(pass=True, score=0.77)


## 2. Example with Patronus client usage

client = Client()

class MyEvaluator(StructureEvaluator):

    def evaluate(self, task_output, gold_answer, **kwargs):
        return task_output.lower().strip() == gold_answer.lower().strip()


client = Client()
client.register_evaluator("my-eval", MyEvaluator())

@traced()
def func_b():
    client.evaluate(evaluators=["my-eval"], task_output=" fOO\t", gold_answer="foo")


func_b()

## Result:
# Span(func_a)
# +- Log(func_name=foo, args=(), output=None)
# +- Span(my-eval)
#    +- Log(log_type=evaluation, body={"task_output": " fOO\t, "gold_answer": "foo"})
#    +- Evaluation(pass=True)


## 3. Ex. with evaluation bundle

@traced()
def func_c():
    client.evaluate(
        evaluators=["my-eval", ("judge", "patronus:fuzzy-match")],
        task_output=" fOO\t", gold_answer="foo"
    )

## Result:
# Span(func_a)
# +- Log(func_name=foo, args=(), output=None)
# +- Span(Evaluation bundle)
#    +- Log(log_type=evaluation, body={"task_output": " fOO\t, "gold_answer": "foo"})
#    +- Span(my-eval)
#       +- Evaluation(pass=True)
#    +- Span(judge-small-2024-08-11:patronus:fuzzy-match)
#       +- Evaluation(pass=True)



import math

import time

import itertools

import logging

import patronus
from patronus import get_api_client
from patronus.api import api_types
from patronus.datasets import Row
from patronus.evals import evaluator, RemoteEvaluator
from patronus.experiments import FuncEvaluatorAdapter, TaskResult, run_experiment

patronus_root_logger = logging.getLogger("patronus")
patronus_root_logger.setLevel(logging.INFO)

console_handler = logging.StreamHandler()
formatter = logging.Formatter(
    fmt='[%(asctime)s] %(levelname)-8s %(name)s: %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S'
)
console_handler.setFormatter(formatter)
patronus_root_logger.addHandler(console_handler)

patronus.init()

log = patronus.get_logger()

api_client = get_api_client()

ANNO_CRITERIA_NAME = "is-correct"

log.info("Listing annotation criteria...")
anno_ls = api_client.list_annotation_criteria_sync().annotation_criteria

anno_exists = False
anno = None
for anno in anno_ls:
    if anno.name == ANNO_CRITERIA_NAME:
        anno_exists = True
        anno = anno

if anno_exists:
    log.info(f"Annotation criteria {ANNO_CRITERIA_NAME!r} exists")
else:
    log.info(f"Annotation criteria does not {ANNO_CRITERIA_NAME!r} exists. Creating...")
    project = api_client.create_project_sync(api_types.CreateProjectRequest(name="Global"))
    anno = api_client.create_annotation_criteria_sync(api_types.CreateAnnotationCriteriaRequest(
        project_id=project.id,
        name=ANNO_CRITERIA_NAME,
        description=None,
        annotation_type=api_types.AnnotationType.binary,
        categories=None,
    )).annotation_criteria

log.info(f"Annotation Criteria ID: {anno.id}")


def my_task(row: Row, **kwargs):
    return f"{row.task_input} World"


# Reference remote Judge Patronus Evaluator with is-concise criteria.
# This evaluator runs remotely on Patronus infrastructure.
is_concise = RemoteEvaluator("judge", "patronus:is-concise")


@evaluator()
def exact_match(row: Row, task_result: TaskResult, **kwargs):
    log.info(f"{task_result.output=}  :: {row.task_output=}")
    return task_result.output == row.task_output


result = run_experiment(
    project_name="Tutorial Project",
    dataset=[
        {
            "task_input": "Hello",
            "gold_answer": "Hello World",
        },
        {
            "task_input": "Bon",
            "gold_answer": "Bon Dia",
        },
    ],
    task=my_task,
    evaluators=[is_concise, FuncEvaluatorAdapter(exact_match)],
)

exp_id = result.experiment.id

log.info("Searching logs...")

logs = []
for c in itertools.count():
    search_logs_resp = api_client.search_logs_sync(
        api_types.SearchLogsRequest(filters=[
            api_types.SearchLogsFilter(
                field="log_attributes['pat.experiment.id']",
                op="eq",
                value=exp_id
            ),
            api_types.SearchLogsFilter(
                field="log_attributes['pat.log.type']",
                op="eq",
                value="eval"
            ),
        ])
    )
    logs = search_logs_resp.logs
    log.info(f"Found {len(search_logs_resp.logs)} logs")
    if len(search_logs_resp.logs) > 0 or c > 10:
        break

    sleep_for = math.pow(c+1, 2)
    time.sleep(sleep_for)
    print(f"Waiting for {sleep_for}s")

log_ids = set()
logs2 = []

for lg in logs:
    if lg.log_attributes["pat.log.id"] not in log_ids:
        log_ids.add(lg.log_attributes["pat.log.id"])
        logs2.append(lg)

log.info(f"deduped logs: {len(logs2)}")

for log in logs2:
    print(f"Log: {log.body}")

    ans = None
    while True:
        user_input = input(">>> Is this correct (yes/no)? ")
        user_input = user_input.strip().lower()
        if user_input == "yes":
            ans = True
        elif user_input == "no":
            ans = False

        if ans is not None:
            break
        else:
            print("Invalid answer.")

        resp = api_client.annotate(api_types.AnnotateRequest(
            annotation_criteria_id=anno.id,
            log_id=log.log_attributes["pat.log.id"],
            value_pass=ans,
        ))
        print(resp)
        print("Annotation given")



from patronus.evals import RemoteEvaluator
from patronus.experiments import run_experiment

dataset = [
    {
        "task_output": "A charming two-story farmhouse perches atop a rolling green hillside, surrounded by mature oak trees and flowering wildflowers. The white clapboard exterior gleams in the late afternoon sunlight, while smoke curls gently from the brick chimney. A winding gravel driveway leads up to the wraparound porch, where a wooden rocking chair offers a perfect vantage point to take in the sweeping valley views below.",
        "task_attachments": [
            {
                "media_type": "image/jpeg",
                "url": "https://helpx-prod.scene7.com/is/image/HelpxProd/upload-content-to-adobe-stock_1408x792-1?$pjpeg$&jpegSize=300&wid=1408",
                "usage_type": "evaluated_model_output"
            }
        ],

    }
]


caption_hallucination = RemoteEvaluator("judge-image", "patronus:caption-hallucination")
caption_describes_primary_object = RemoteEvaluator("judge-image", "patronus:caption-describes-primary-object")
caption_describes_non_primary_objects = RemoteEvaluator("judge-image",
                                                                "patronus:caption-describes-non-primary-objects")
caption_mentions_primary_object_location = RemoteEvaluator("judge-image",
                                                                   "patronus:caption-mentions-primary-object-location")
caption_hallucination_strict = RemoteEvaluator("judge-image", "patronus:caption-hallucination-strict")

results = run_experiment(
    project_name="Multimodal Experiments",
    experiment_name="GPT-4o Image Captioning",
    dataset=dataset,
    evaluators=[caption_hallucination],
    tags={"dataset_key": "8233883", "model_key": "gpt-4o",
          "system_prompt": "This is an image for which you need to generate a caption."}
)